@@ -45,7 +45,7 @@ double test_gemm(int M, int N, int K, int transa)
     printf("b.ld() = %i\n", b.ld());
     printf("c.ld() = %i\n", c.ld());
     utils::timer t1("gemm_only"); 
-    linalg<CPU>::gemm(transa, 0, M, N, K, a.at(memory_t::host), a.ld(), b.at(memory_t::host), b.ld(), c.at(memory_t::host), c.ld());
+    linalg<device_t::CPU>::gemm(transa, 0, M, N, K, a.at(memory_t::host), a.ld(), b.at(memory_t::host), b.ld(), c.at(memory_t::host), c.ld());
     double tval = t1.stop();
     double perf = nop_gemm * 1e-9 * M * N * K / tval;
     printf("execution time (sec) : %12.6f\n", tval);
@@ -97,7 +97,7 @@ double test_pgemm(int M, int N, int K, int nrow, int ncol, int transa, int n, in
     utils::timer t1("gemm_only");
     gemm_type one = 1;
     gemm_type zero = 0;
-    linalg<CPU>::gemm(transa, 0, M, N - n, K, one, a, 0, 0, b, 0, n, zero, c, 0, 0);
+    linalg<device_t::CPU>::gemm(transa, 0, M, N - n, K, one, a, 0, 0, b, 0, n, zero, c, 0, 0);
     //== #ifdef _GPU_
     //== cuda_device_synchronize();
     //== #endif
@@ -143,7 +143,7 @@ double test_pgemm(int M, int N, int K, int nrow, int ncol, int transa, int n, in
 //    sirius::Timer t1("gemm_only", comm); 
 //    gemm_type one = 1;
 //    gemm_type zero = 0;
-//    linalg<CPU>::gemm(2, 0, M, M, K, one, a, 0, 0, a, 0, 0, zero, c, 0, 0);
+//    linalg<device_t::CPU>::gemm(2, 0, M, M, K, one, a, 0, 0, a, 0, 0, zero, c, 0, 0);
 //    #ifdef _GPU_
 //    cuda_device_synchronize();
 //    #endif
