@@ -26,9 +26,10 @@ inline void Band::diag_full_potential_first_variation_exact(K_point& kp, Hamilto
 {
     PROFILE("sirius::Band::diag_fv_exact");
 
-    auto                    mem_type = (ctx_.gen_evp_solver_type() == ev_solver_t::magma) ? memory_t::host_pinned : memory_t::host;
-    int                     ngklo    = kp.gklo_basis_size();
-    int                     bs       = ctx_.cyclic_block_size();
+    auto mem_type = (ctx_.gen_evp_solver_type() == ev_solver_t::magma) ? memory_t::host_pinned : memory_t::host;
+    int  ngklo    = kp.gklo_basis_size();
+    int  bs       = ctx_.cyclic_block_size();
+
     dmatrix<double_complex> h(ngklo, ngklo, ctx_.blacs_grid(), bs, bs, mem_type);
     dmatrix<double_complex> o(ngklo, ngklo, ctx_.blacs_grid(), bs, bs, mem_type);
 
@@ -38,13 +39,13 @@ inline void Band::diag_full_potential_first_variation_exact(K_point& kp, Hamilto
 
     /* setup Hamiltonian and overlap */
     switch (ctx_.processing_unit()) {
-        case CPU: {
-            hamiltonian__.set_fv_h_o<CPU, electronic_structure_method_t::full_potential_lapwlo>(&kp, h, o);
+        case device_t::CPU: {
+            hamiltonian__.set_fv_h_o<device_t::CPU, electronic_structure_method_t::full_potential_lapwlo>(&kp, h, o);
             break;
         }
 #ifdef __GPU
-        case GPU: {
-            hamiltonian__.set_fv_h_o<GPU, electronic_structure_method_t::full_potential_lapwlo>(&kp, h, o);
+        case device_t::GPU: {
+            hamiltonian__.set_fv_h_o<device_t::GPU, electronic_structure_method_t::full_potential_lapwlo>(&kp, h, o);
             break;
         }
 #endif
@@ -118,15 +119,17 @@ inline void Band::diag_full_potential_first_variation_exact(K_point& kp, Hamilto
                            [this](int ia) { return unit_cell_.atom(ia).mt_lo_basis_size(); }, ctx_.num_fv_states(),
                            ctx_.preferred_memory_t(), 1);
         if (ctx_.processing_unit() == device_t::GPU) {
-            kp.fv_eigen_vectors_slab().allocate(spin_idx(0), memory_t::device);
-            kp.fv_eigen_vectors_slab().copy_to(spin_idx(0), memory_t::device, 0, ctx_.num_fv_states());
-            ofv.allocate(spin_idx(0), memory_t::device);
+            kp.fv_eigen_vectors_slab().allocate(spin_range(0), memory_t::device);
+            kp.fv_eigen_vectors_slab().copy_to(spin_range(0), memory_t::device, 0, ctx_.num_fv_states());
+            ofv.allocate(spin_range(0), memory_t::device);
         }
 
+        hamiltonian__.local_op().prepare(kp.gkvec_partition());
         hamiltonian__.apply_fv_h_o(&kp, false, false, 0, ctx_.num_fv_states(), kp.fv_eigen_vectors_slab(), nullptr, &ofv);
+        hamiltonian__.local_op().dismiss();
 
         if (ctx_.processing_unit() == device_t::GPU) {
-            kp.fv_eigen_vectors_slab().deallocate(spin_idx(0), memory_t::device);
+            kp.fv_eigen_vectors_slab().deallocate(spin_range(0), memory_t::device);
         }
 
         std::vector<double> norm(ctx_.num_fv_states(), 0);
@@ -207,7 +210,7 @@ inline void Band::get_singular_components(K_point& kp__, Hamiltonian& H__) const
         diag1[ig]  = 1;
     }
 
-    if (ctx_.processing_unit() == GPU) {
+    if (ctx_.processing_unit() == device_t::GPU) {
         o_diag.allocate(memory_t::device).copy_to(memory_t::device);
         diag1.allocate(memory_t::device).copy_to(memory_t::device);
     }
@@ -349,7 +352,8 @@ inline void Band::get_singular_components(K_point& kp__, Hamiltonian& H__) const
         /* don't compute residuals on last iteration */
         if (k != itso.num_steps_ - 1) {
             /* get new preconditionined residuals, and also opsi and psi as a by-product */
-            n = residuals(&kp__, 0, N, ncomp, eval, eval_old, evec, ophi, phi, opsi, psi, res, o_diag, diag1);
+            n = residuals(&kp__, 0, N, ncomp, eval, eval_old, evec, ophi, phi, opsi, psi, res, o_diag, diag1,
+                          itso.energy_tolerance_, itso.residual_tolerance_);
         }
 
         /* check if we run out of variational space or eigen-vectors are converged or it's a last iteration */
@@ -488,19 +492,19 @@ inline void Band::diag_full_potential_first_variation_davidson(K_point& kp__, Ha
     }
 
     if (ctx_.processing_unit() == device_t::GPU) {
-        psi.allocate(spin_idx(0), memory_t::device);
-        psi.copy_to(spin_idx(0), memory_t::device, 0, num_bands);
+        psi.allocate(spin_range(0), memory_t::device);
+        psi.copy_to(spin_range(0), memory_t::device, 0, num_bands);
 
-        phi.allocate(spin_idx(0), memory_t::device);
-        phi.copy_to(spin_idx(0), memory_t::device, 0, nlo + ncomp);
+        phi.allocate(spin_range(0), memory_t::device);
+        phi.copy_to(spin_range(0), memory_t::device, 0, nlo + ncomp);
 
-        res.allocate(spin_idx(0), memory_t::device);
+        res.allocate(spin_range(0), memory_t::device);
 
-        hphi.allocate(spin_idx(0), memory_t::device);
-        ophi.allocate(spin_idx(0), memory_t::device);
+        hphi.allocate(spin_range(0), memory_t::device);
+        ophi.allocate(spin_range(0), memory_t::device);
 
-        hpsi.allocate(spin_idx(0), memory_t::device);
-        opsi.allocate(spin_idx(0), memory_t::device);
+        hpsi.allocate(spin_range(0), memory_t::device);
+        opsi.allocate(spin_range(0), memory_t::device);
 
         if (ctx_.blacs_grid().comm().size() == 1) {
             evec.allocate(memory_t::device);
@@ -580,7 +584,8 @@ inline void Band::diag_full_potential_first_variation_davidson(K_point& kp__, Ha
         /* don't compute residuals on last iteration */
         if (k != itso.num_steps_ - 1) {
             /* get new preconditionined residuals, and also hpsi and opsi as a by-product */
-            n = residuals(&kp__, 0, N, num_bands, eval, eval_old, evec, hphi, ophi, hpsi, opsi, res, h_diag, o_diag);
+            n = residuals(&kp__, 0, N, num_bands, eval, eval_old, evec, hphi, ophi, hpsi, opsi, res, h_diag, o_diag,
+                          itso.energy_tolerance_, itso.residual_tolerance_);
         }
 
         /* check if we run out of variational space or eigen-vectors are converged or it's a last iteration */
@@ -615,7 +620,7 @@ inline void Band::diag_full_potential_first_variation_davidson(K_point& kp__, Ha
     if (ctx_.processing_unit() == device_t::GPU) {
         psi.pw_coeffs(0).copy_to(memory_t::host, 0, num_bands);
         psi.mt_coeffs(0).copy_to(memory_t::host, 0, num_bands);
-        psi.deallocate(spin_idx(0), memory_t::device);
+        psi.deallocate(spin_range(0), memory_t::device);
     }
     kp__.set_fv_eigen_values(&eval[0]);
 }
@@ -668,17 +673,19 @@ inline void Band::diag_full_potential_second_variation(K_point& kp__, Hamiltonia
     //==     }
     //== }
 
-    //== if (ctx_.so_correction()) apply_so_correction(kp->fv_states_col(), hpsi);
+    if (ctx_.so_correction()) {
+        hamiltonian__.apply_so_correction(&kp__, kp__.fv_states(), hpsi);
+    }
 
     int nfv = ctx_.num_fv_states();
     int bs  = ctx_.cyclic_block_size();
 
     if (ctx_.processing_unit() == device_t::GPU) {
-        kp__.fv_states().allocate(spin_idx(0), memory_t::device);
-        kp__.fv_states().copy_to(spin_idx(0), memory_t::device, 0, nfv);
+        kp__.fv_states().allocate(spin_range(0), memory_t::device);
+        kp__.fv_states().copy_to(spin_range(0), memory_t::device, 0, nfv);
         for (int i = 0; i < ctx_.num_mag_comp(); i++) {
-            hpsi[i].allocate(spin_idx(0), memory_t::device);
-            hpsi[i].copy_to(spin_idx(0), memory_t::device, 0, nfv);
+            hpsi[i].allocate(spin_range(0), memory_t::device);
+            hpsi[i].copy_to(spin_range(0), memory_t::device, 0, nfv);
         }
     }
 
@@ -699,12 +706,12 @@ inline void Band::diag_full_potential_second_variation(K_point& kp__, Hamiltonia
     linalg_t la{linalg_t::blas};
     if (ctx_.processing_unit() == device_t::GPU) {
         mem = memory_t::device;
-        la = linalg_t::cublas;
+        la = linalg_t::gpublas;
     }
 
     if (ctx_.num_mag_dims() != 3) {
         dmatrix<double_complex> h(nfv, nfv, ctx_.blacs_grid(), bs, bs);
-        if (ctx_.blacs_grid().comm().size() == 1 && ctx_.processing_unit() == GPU) {
+        if (ctx_.blacs_grid().comm().size() == 1 && ctx_.processing_unit() == device_t::GPU) {
             h.allocate(memory_t::device);
         }
         /* perform one or two consecutive diagonalizations */
@@ -724,9 +731,9 @@ inline void Band::diag_full_potential_second_variation(K_point& kp__, Hamiltonia
             std_solver.solve(nfv, nfv, h, &band_energies(0, ispn), kp__.sv_eigen_vectors(ispn));
         }
     } else {
-        int                     nb = ctx_.num_bands();
+        int nb = ctx_.num_bands();
         dmatrix<double_complex> h(nb, nb, ctx_.blacs_grid(), bs, bs);
-        if (ctx_.blacs_grid().comm().size() == 1 && ctx_.processing_unit() == GPU) {
+        if (ctx_.blacs_grid().comm().size() == 1 && ctx_.processing_unit() == device_t::GPU) {
             h.allocate(memory_t::device);
         }
         /* compute <wf_i | h * wf_j> for up-up block */
@@ -759,9 +766,9 @@ inline void Band::diag_full_potential_second_variation(K_point& kp__, Hamiltonia
     }
 
     if (ctx_.processing_unit() == device_t::GPU) {
-        kp__.fv_states().deallocate(spin_idx(0), memory_t::device);
+        kp__.fv_states().deallocate(spin_range(0), memory_t::device);
         for (int i = 0; i < ctx_.num_mag_comp(); i++) {
-            hpsi[i].deallocate(spin_idx(0), memory_t::device);
+            hpsi[i].deallocate(spin_range(0), memory_t::device);
         }
     }
     for (int ispn = 0; ispn < ctx_.num_spin_dims(); ispn++) {
