@@ -1,20 +1,20 @@
 // Copyright (c) 2013-2017 Anton Kozhevnikov, Thomas Schulthess
 // All rights reserved.
 //
-// Redistribution and use in source and binary forms, with or without modification, are permitted provided that 
+// Redistribution and use in source and binary forms, with or without modification, are permitted provided that
 // the following conditions are met:
 //
-// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the 
+// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 //    following disclaimer.
-// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
+// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
 //    and the following disclaimer in the documentation and/or other materials provided with the distribution.
 //
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
-// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
-// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR 
-// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
-// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
-// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
+// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
+// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 // OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 /** \file beta_projectors_base.hpp
@@ -24,11 +24,13 @@
 
 #ifndef __BETA_PROJECTORS_BASE_HPP__
 #define __BETA_PROJECTORS_BASE_HPP__
+
 #include "utils/env.hpp"
+#include "simulation_context.hpp"
 
 namespace sirius {
 
-#ifdef __GPU
+#if defined(__GPU)
 extern "C" void create_beta_gk_gpu(int                   num_atoms,
                                    int                   num_gkvec,
                                    int const*            beta_desc,
@@ -38,21 +40,26 @@ extern "C" void create_beta_gk_gpu(int                   num_atoms,
                                    double_complex*       beta_gk);
 #endif
 
-enum beta_desc_idx
+/// Named index of a descriptor of beta-projectors. The same order is used by the GPU kernel.
+enum class beta_desc_idx
 {
+    /// Number of beta-functions for this atom.
     nbf      = 0,
+    /// Offset of beta-projectors in this chunk.
     offset   = 1,
+    /// Offset of beta-projectors in the array for atom types.
     offset_t = 2,
+    /// Global index of atom.
     ia       = 3
 };
 
-struct beta_chunk_t 
+struct beta_chunk_t
 {
     /// Number of beta-projectors in the current chunk.
     int num_beta_;
     /// Number of atoms in the current chunk.
     int num_atoms_;
-    /// Offset in the global index of beta projectors. 
+    /// Offset in the global index of beta projectors.
     int offset_;
     /// Descriptor of block of beta-projectors for an atom.
     mdarray<int, 2> desc_;
@@ -128,13 +135,13 @@ class Beta_projectors_base
                     beta_chunks_[ib].atom_pos_(x, i) = pos[x];
                 }
                 /* number of beta functions for atom */
-                beta_chunks_[ib].desc_(beta_desc_idx::nbf, i) = type.mt_basis_size();
+                beta_chunks_[ib].desc_(static_cast<int>(beta_desc_idx::nbf), i) = type.mt_basis_size();
                 /* offset in beta_gk*/
-                beta_chunks_[ib].desc_(beta_desc_idx::offset, i) = num_beta;
+                beta_chunks_[ib].desc_(static_cast<int>(beta_desc_idx::offset), i) = num_beta;
                 /* offset in beta_gk_t */
-                beta_chunks_[ib].desc_(beta_desc_idx::offset_t, i) = type.offset_lo();
+                beta_chunks_[ib].desc_(static_cast<int>(beta_desc_idx::offset_t), i) = type.offset_lo();
                 /* global index of atom */
-                beta_chunks_[ib].desc_(beta_desc_idx::ia, i) = ia;
+                beta_chunks_[ib].desc_(static_cast<int>(beta_desc_idx::ia), i) = ia;
 
                 num_beta += type.mt_basis_size();
             }
@@ -219,7 +226,7 @@ class Beta_projectors_base
 
     matrix<double_complex> pw_coeffs_t(int j__)
     {
-        return std::move(matrix<double_complex>(&pw_coeffs_t_(0, 0, j__), num_gkvec_loc(), num_beta_t()));
+        return matrix<double_complex>(&pw_coeffs_t_(0, 0, j__), num_gkvec_loc(), num_beta_t());
     }
 
     /// Plane wave coefficients of |beta> projectors for a chunk of atoms.
@@ -284,14 +291,14 @@ class Beta_projectors_base
             case device_t::CPU: break;
         }
 
-        return std::move(beta_phi);
+        return beta_phi;
     }
 
     /// Generate beta-projectors for a chunk of atoms.
     /** Beta-projectors are always generated and stored in the memory of a processing unit.
      *
      *  \param [in] ichunk Index of a chunk of atoms for which beta-projectors are generated.
-     *  \param [in] j index of the component (up to 9 components are used for the strain derivative) 
+     *  \param [in] j index of the component (up to 9 components are used for the strain derivative)
      */
     void generate(int ichunk__, int j__)
     {
@@ -299,9 +306,9 @@ class Beta_projectors_base
 
         switch (ctx_.processing_unit()) {
             case device_t::CPU: {
-                #pragma omp for
+                #pragma omp parallel for
                 for (int i = 0; i < chunk(ichunk__).num_atoms_; i++) {
-                    int ia = chunk(ichunk__).desc_(beta_desc_idx::ia, i);
+                    int ia = chunk(ichunk__).desc_(static_cast<int>(beta_desc_idx::ia), i);
 
                     double phase = twopi * dot(gkvec_.vk(), ctx_.unit_cell().atom(ia).position());
                     double_complex phase_k = std::exp(double_complex(0.0, phase));
@@ -312,17 +319,17 @@ class Beta_projectors_base
                         /* total phase e^{i(G+k)r_{\alpha}} */
                         phase_gk[igk_loc] = std::conj(ctx_.gvec_phase_factor(G, ia) * phase_k);
                     }
-                    for (int xi = 0; xi < chunk(ichunk__).desc_(beta_desc_idx::nbf, i); xi++) {
+                    for (int xi = 0; xi < chunk(ichunk__).desc_(static_cast<int>(beta_desc_idx::nbf), i); xi++) {
                         for (int igk_loc = 0; igk_loc < num_gkvec_loc(); igk_loc++) {
-                            pw_coeffs_a_(igk_loc, chunk(ichunk__).desc_(beta_desc_idx::offset, i) + xi) = 
-                                pw_coeffs_t_(igk_loc, chunk(ichunk__).desc_(beta_desc_idx::offset_t, i) + xi, j__) * phase_gk[igk_loc];
+                            pw_coeffs_a_(igk_loc, chunk(ichunk__).desc_(static_cast<int>(beta_desc_idx::offset), i) + xi) =
+                                pw_coeffs_t_(igk_loc, chunk(ichunk__).desc_(static_cast<int>(beta_desc_idx::offset_t), i) + xi, j__) * phase_gk[igk_loc];
                         }
                     }
                 }
                 break;
             }
             case device_t::GPU: {
-#ifdef __GPU
+#if defined(__GPU)
                 auto& desc = chunk(ichunk__).desc_;
                 create_beta_gk_gpu(chunk(ichunk__).num_atoms_,
                                    num_gkvec_loc(),
@@ -335,11 +342,11 @@ class Beta_projectors_base
                 /* wave-functions are on CPU but the beta-projectors are on GPU */
                 if (gkvec_.comm().rank() == 0 && is_host_memory(ctx_.preferred_memory_t())) {
                     /* make beta-projectors for G=0 on the CPU */
-                    #pragma omp for schedule(static)
+                    #pragma omp parallel for schedule(static)
                     for (int i = 0; i < chunk(ichunk__).num_atoms_; i++) {
-                        for (int xi = 0; xi < chunk(ichunk__).desc_(beta_desc_idx::nbf, i); xi++) {
-                            pw_coeffs_a_g0_(chunk(ichunk__).desc_(beta_desc_idx::offset, i) + xi) =
-                                pw_coeffs_t_(0, chunk(ichunk__).desc_(beta_desc_idx::offset_t, i) + xi, j__);
+                        for (int xi = 0; xi < chunk(ichunk__).desc_(static_cast<int>(beta_desc_idx::nbf), i); xi++) {
+                            pw_coeffs_a_g0_(chunk(ichunk__).desc_(static_cast<int>(beta_desc_idx::offset), i) + xi) =
+                                pw_coeffs_t_(0, chunk(ichunk__).desc_(static_cast<int>(beta_desc_idx::offset_t), i) + xi, j__);
                         }
                     }
                 }
@@ -360,8 +367,7 @@ class Beta_projectors_base
                 break;
             }
             case device_t::GPU: {
-                pw_coeffs_a_ = matrix<double_complex>(nullptr, num_gkvec_loc(), max_num_beta());
-                pw_coeffs_a_.allocate(ctx_.mem_pool(memory_t::device));
+                pw_coeffs_a_ = matrix<double_complex>(ctx_.mem_pool(memory_t::device), num_gkvec_loc(), max_num_beta());
                 pw_coeffs_a_g0_ = mdarray<double_complex, 1>(ctx_.mem_pool(memory_t::host), max_num_beta());
                 pw_coeffs_a_g0_.allocate(ctx_.mem_pool(memory_t::device));
                 break;
@@ -377,7 +383,7 @@ class Beta_projectors_base
     {
         PROFILE("sirius::Beta_projectors_base::dismiss");
 
-        if (ctx_.processing_unit() == GPU && reallocate_pw_coeffs_t_on_gpu_) {
+        if (ctx_.processing_unit() == device_t::GPU && reallocate_pw_coeffs_t_on_gpu_) {
             pw_coeffs_t_.deallocate(memory_t::device);
         }
         pw_coeffs_a_.deallocate(memory_t::device);
@@ -423,7 +429,7 @@ inline void Beta_projectors_base::local_inner_aux<double_complex>(double_complex
     auto pp = utils::get_env<int>("SIRIUS_PRINT_PERFORMANCE");
     if (pp && gkvec_.comm().rank() == 0) {
 #ifdef __GPU
-        if (ctx_.blas_linalg_t() == linalg_t::cublas) {
+        if (ctx_.blas_linalg_t() == linalg_t::gpublas) {
             acc::sync_stream(stream_id(-1));
         }
 #endif
@@ -453,7 +459,7 @@ inline void Beta_projectors_base::local_inner_aux<double>(double* beta_pw_coeffs
         linalg_t la{linalg_t::none};
         /* both wave-functions and beta-projectors are on GPU */
         if (is_device_memory(ctx_.preferred_memory_t())) {
-            la = linalg_t::cublas;
+            la = linalg_t::gpublas;
         } else { /* wave-functions are on CPU but the beta-projectors are in the memory of main device */
             la = linalg_t::blas;
             switch (ctx_.processing_unit()) {
