@@ -139,9 +139,9 @@ class Density : public Field4D
 
         int ia{-1};
 
-        /// ae and ps local unified densities+magnetization
-        std::vector<Spheric_function<spectral, double>> ae_density_;
-        std::vector<Spheric_function<spectral, double>> ps_density_;
+        /* AE and PS local unified densities + magnetization */
+        std::vector<Spheric_function<function_domain_t::spectral, double>> ae_density_;
+        std::vector<Spheric_function<function_domain_t::spectral, double>> ps_density_;
     };
 
     std::vector<paw_density_data_t> paw_density_data_;
@@ -165,20 +165,11 @@ class Density : public Field4D
     /// Fast mapping between composite lm index and corresponding orbital quantum number.
     std::vector<int> l_by_lm_;
 
-    /// High-frequency mixer for the pseudopotential density mixing.
-    std::unique_ptr<Mixer<double_complex>> hf_mixer_{nullptr};
-
     /// Low-frequency mixer for the pseudopotential density mixing.
-    std::unique_ptr<Mixer<double_complex>> lf_mixer_{nullptr};
+    std::unique_ptr<Mixer<double_complex>> mixer_{nullptr};
 
-    /// List of local low-fequency G-vectors.
-    std::vector<int> lf_gvec_;
-
-    /// List of local high-fequency G-vectors.
-    std::vector<int> hf_gvec_;
-
-    /// Weights of local low-frequency G-vectors.
-    std::vector<double> lf_gvec_weights_;
+    /// Weights of G-vectors for the mixer
+    std::vector<double> gvec_mixer_weights_;
 
     /// Allocate PAW data.
     void init_paw();
@@ -310,6 +301,8 @@ class Density : public Field4D
         : Field4D(ctx__, ctx__.lmmax_rho())
         , unit_cell_(ctx_.unit_cell())
     {
+        PROFILE("sirius::Density");
+
         if (!ctx_.initialized()) {
             TERMINATE("Simulation_context is not initialized");
         }
@@ -341,15 +334,10 @@ class Density : public Field4D
         for (int igloc = 0; igloc < ctx_.gvec().count(); igloc++) {
             int  ig = ctx_.gvec().offset() + igloc;
             auto gv = ctx_.gvec().gvec_cart<index_domain_t::local>(igloc);
-            if (gv.length() <= 2 * ctx_.gk_cutoff()) {
-                lf_gvec_.push_back(igloc);
-                if (ig) {
-                    lf_gvec_weights_.push_back(fourpi * unit_cell_.omega() / std::pow(gv.length(), 2));
-                } else {
-                    lf_gvec_weights_.push_back(0);
-                }
+            if (ig) {
+                gvec_mixer_weights_.push_back(fourpi * unit_cell_.omega() / std::pow(gv.length(), 2));
             } else {
-                hf_gvec_.push_back(igloc);
+                gvec_mixer_weights_.push_back(0);
             }
         }
 
@@ -397,9 +385,7 @@ class Density : public Field4D
 
     inline void normalize()
     {
-        std::vector<double> nel_mt;
-        double              nel_it;
-        double nel   = rho().integrate(nel_mt, nel_it);
+        double nel = std::get<0>(rho().integrate());
         double scale = unit_cell_.num_electrons() / nel;
 
         /* renormalize interstitial part */
@@ -422,9 +408,7 @@ class Density : public Field4D
     {
         double nel{0};
         if (ctx_.full_potential()) {
-            std::vector<double> nel_mt;
-            double              nel_it;
-            nel = rho().integrate(nel_mt, nel_it);
+            nel = std::get<0>(rho().integrate());
         } else {
             nel = rho().f_0().real() * unit_cell_.omega();
         }
@@ -539,12 +523,12 @@ class Density : public Field4D
 
         switch (ctx_.processing_unit()) {
             case device_t::CPU: {
-                generate_rho_aug<CPU>(rho_aug);
+                generate_rho_aug<device_t::CPU>(rho_aug);
                 break;
             }
             case device_t::GPU: {
                 rho_aug.allocate(memory_t::device);
-                generate_rho_aug<GPU>(rho_aug);
+                generate_rho_aug<device_t::GPU>(rho_aug);
                 break;
             }
         }
@@ -802,7 +786,7 @@ class Density : public Field4D
         return this->vector(i);
     }
 
-    Spheric_function<spectral, double> const& density_mt(int ialoc) const
+    Spheric_function<function_domain_t::spectral, double> const& density_mt(int ialoc) const
     {
         return rho().f_mt(ialoc);
     }
@@ -810,12 +794,12 @@ class Density : public Field4D
     /// Generate \f$ n_1 \f$  and \f$ \tilde{n}_1 \f$ in lm components.
     void generate_paw_loc_density();
 
-    std::vector<Spheric_function<spectral, double>> const& ae_paw_atom_density(int spl_paw_ind) const
+    std::vector<Spheric_function<function_domain_t::spectral, double>> const& ae_paw_atom_density(int spl_paw_ind) const
     {
         return paw_density_data_[spl_paw_ind].ae_density_;
     }
 
-    std::vector<Spheric_function<spectral, double>> const& ps_paw_atom_density(int spl_paw_ind) const
+    std::vector<Spheric_function<function_domain_t::spectral, double>> const& ps_paw_atom_density(int spl_paw_ind) const
     {
         return paw_density_data_[spl_paw_ind].ps_density_;
     }
@@ -825,35 +809,22 @@ class Density : public Field4D
         if (ctx_.full_potential()) {
             Field4D::mixer_input();
         } else {
-            int ld = static_cast<int>(hf_gvec_.size());
-            if (hf_mixer_) {
-                /* input high-frequency components */
-                for (int j = 0; j < ctx_.num_mag_dims() + 1; j++) {
-                    for (int i = 0; i < static_cast<int>(hf_gvec_.size()); i++) {
-                        int igloc = hf_gvec_[i];
-                        hf_mixer_->input_local(i + j * ld, component(j).f_pw_local(igloc));
-                    }
-                }
-            }
-
-            ld = static_cast<int>(lf_gvec_.size());
-            /* input low-frequency components */
+            int ngv = ctx_.gvec().count();
             for (int j = 0; j < ctx_.num_mag_dims() + 1; j++) {
                 if (j == 0) {
-                    for (int i = 0; i < static_cast<int>(lf_gvec_.size()); i++) {
-                        int igloc = lf_gvec_[i];
-                        lf_mixer_->input_local(i + j * ld, component(j).f_pw_local(igloc), lf_gvec_weights_[i]);
+                    for (int igloc = 0; igloc < ngv; igloc++) {
+                        //mixer_->input_local(igloc + j * ngv, component(j).f_pw_local(igloc), gvec_mixer_weights_[igloc]);
+                        mixer_->input_local(igloc + j * ngv, component(j).f_pw_local(igloc));
                     }
                 } else {
-                    for (int i = 0; i < static_cast<int>(lf_gvec_.size()); i++) {
-                        int igloc = lf_gvec_[i];
-                        lf_mixer_->input_local(i + j * ld, component(j).f_pw_local(igloc));
+                    for (int igloc = 0; igloc < ngv; igloc++) {
+                        mixer_->input_local(igloc + j * ngv, component(j).f_pw_local(igloc));
                     }
                 }
             }
             /* input commonly shared data */
             for (int i = 0; i < static_cast<int>(density_matrix_.size()); i++) {
-                lf_mixer_->input_shared(i, density_matrix_[i], 0);
+                mixer_->input_shared(i, density_matrix_[i], 0);
             }
         }
     }
@@ -863,28 +834,16 @@ class Density : public Field4D
         if (ctx_.full_potential()) {
             Field4D::mixer_output();
         } else {
-            int ld = static_cast<int>(hf_gvec_.size());
-            if (hf_mixer_) {
-                /* get high-frequency components */
-                for (int j = 0; j < ctx_.num_mag_dims() + 1; j++) {
-                    for (int i = 0; i < static_cast<int>(hf_gvec_.size()); i++) {
-                        int igloc                      = hf_gvec_[i];
-                        component(j).f_pw_local(igloc) = hf_mixer_->output_local(i + j * ld);
-                    }
-                }
-            }
-
-            ld = static_cast<int>(lf_gvec_.size());
+            int ngv = ctx_.gvec().count();
             /* get low-frequency components */
             for (int j = 0; j < ctx_.num_mag_dims() + 1; j++) {
-                for (int i = 0; i < static_cast<int>(lf_gvec_.size()); i++) {
-                    int igloc                      = lf_gvec_[i];
-                    component(j).f_pw_local(igloc) = lf_mixer_->output_local(i + j * ld);
+                for (int igloc = 0; igloc < ngv; igloc++) {
+                    component(j).f_pw_local(igloc) = mixer_->output_local(igloc + j * ngv);
                 }
             }
 
             for (int i = 0; i < static_cast<int>(density_matrix_.size()); i++) {
-                density_matrix_[i] = lf_mixer_->output_shared(i);
+                density_matrix_[i] = mixer_->output_shared(i);
             }
         }
     }
@@ -894,25 +853,11 @@ class Density : public Field4D
         if (ctx_.full_potential()) {
             Field4D::mixer_init(mixer_cfg__);
         } else {
-            auto conf1  = mixer_cfg__;
-            conf1.type_ = "linear";
-            if (hf_gvec_.size() > 0) {
-                hf_mixer_   = Mixer_factory<double_complex>(0,
-                                                            static_cast<int>(hf_gvec_.size() * (1 + ctx_.num_mag_dims())),
-                                                            conf1,
-                                                            ctx_.comm());
-            }
-
-            lf_mixer_ = Mixer_factory<double_complex>(static_cast<int>(density_matrix_.size()),
-                                                      static_cast<int>(lf_gvec_.size() * (1 + ctx_.num_mag_dims())),
-                                                      mixer_cfg__,
-                                                      ctx_.comm());
+            mixer_ = Mixer_factory<double_complex>(static_cast<int>(density_matrix_.size()),
+                                                   ctx_.gvec().count() * (1 + ctx_.num_mag_dims()),
+                                                   mixer_cfg__, ctx_.comm());
             mixer_input();
-            lf_mixer_->initialize();
-
-            if (hf_mixer_) {
-                hf_mixer_->initialize();
-            }
+            mixer_->initialize();
         }
     }
 
@@ -928,20 +873,17 @@ class Density : public Field4D
         } else {
             /* mix in G-space in case of PP */
             mixer_input();
-            rms = lf_mixer_->mix(ctx_.settings().mixer_rss_min_);
-            if (hf_mixer_) {
-                rms += hf_mixer_->mix(ctx_.settings().mixer_rss_min_);
-            }
+            rms = mixer_->mix(ctx_.settings().mixer_rss_min_);
             mixer_output();
         }
 
         return rms;
     }
 
-    inline double dr2() const
-    {
-        return lf_mixer_->rss();
-    }
+    //inline double dr2() const
+    //{
+    //    return mixer_->rss();
+    //}
 
     mdarray<double_complex, 4> const& density_matrix() const
     {
@@ -986,7 +928,7 @@ class Density : public Field4D
                 }
             }
         }
-        return std::move(dm);
+        return dm;
     }
 
     /// Calculate approximate atomic magnetic moments in case of PP-PW.
@@ -1014,7 +956,7 @@ class Density : public Field4D
             }
         }
         ctx_.fft().comm().allreduce(&mmom(0, 0), static_cast<int>(mmom.size()));
-        return std::move(mmom);
+        return mmom;
     }
 
     /// Symmetrize density matrix.
@@ -1060,11 +1002,13 @@ class Density : public Field4D
      *  \f]
      */
     void symmetrize_density_matrix();
-
+    void symmetrize_density_matrix_bis();
     void symmetrize()
     {
         Field4D::symmetrize(&rho(), &magnetization(0), &magnetization(1), &magnetization(2));
     }
+
+#include "Symmetrize.hpp"
 };
 
 #include "initial_density.hpp"
