@@ -26,6 +26,7 @@
 #define __LOCAL_OPERATOR_HPP__
 
 #include "Potential/potential.hpp"
+#include "../SDDK/GPU/acc.hpp"
 
 #ifdef __GPU
 extern "C" void mul_by_veff_gpu(int ispn__, int size__, double* const* veff__, double_complex* buf__);
@@ -90,6 +91,8 @@ class Local_operator
         , gvec_coarse_p_(gvec_coarse_p__)
 
     {
+        PROFILE("sirius::Local_operator");
+
         for (int j = 0; j < ctx_.num_mag_dims() + 1; j++) {
             veff_vec_[j] = Smooth_periodic_function<double>(fft_coarse__, gvec_coarse_p__);
             for (int ir = 0; ir < fft_coarse_.local_size(); ir++) {
@@ -168,7 +171,7 @@ class Local_operator
             /* release FFT driver */
             fft_coarse_.dismiss();
 
-            if (fft_coarse_.pu() == GPU) {
+            if (fft_coarse_.pu() == device_t::GPU) {
                 for (int j = 0; j < ctx_.num_mag_dims() + 1; j++) {
                     veff_vec_[j].f_rg().allocate(memory_t::device).copy_to(memory_t::device);
                 }
@@ -215,7 +218,7 @@ class Local_operator
             }
 
             /* copy veff to device */
-            if (fft_coarse_.pu() == GPU) {
+            if (fft_coarse_.pu() == device_t::GPU) {
                 for (int j = 0; j < ctx_.num_mag_dims() + 1; j++) {
                     veff_vec_[j].f_rg().allocate(memory_t::device).copy_to(memory_t::device);
                 }
@@ -272,7 +275,7 @@ class Local_operator
     /// Cleanup the local operator.
     inline void dismiss()
     {
-        if (fft_coarse_.pu() == GPU) {
+        if (fft_coarse_.pu() == device_t::GPU) {
             for (int j = 0; j < ctx_.num_mag_dims() + 1; j++) {
                 veff_vec_[j].f_rg().deallocate(memory_t::device);
             }
@@ -559,7 +562,7 @@ class Local_operator
         };
 
         /* store the resulting hphi
-           spin block (ispn_block) is used as a bit mask: 
+           spin block (ispn_block) is used as a bit mask:
             - first bit: spin component which is updated
             - second bit: add or not kinetic energy term */
         auto add_to_hphi = [&](int ispn_block, bool gamma = false) {
@@ -635,7 +638,7 @@ class Local_operator
 
         int first{0};
         /* If G-vectors are reduced, wave-functions are real and we can transform two of them at once.
-           Non-collinear case is not treated here because nc wave-functions are complex and G+k vectors 
+           Non-collinear case is not treated here because nc wave-functions are complex and G+k vectors
            can't be reduced. In this case input spin index can only be 0 or 1. */
         if (gkvec_p_->gvec().reduced()) {
             int npairs = num_wf_loc / 2;
@@ -753,11 +756,11 @@ class Local_operator
                 mul_by_veff(buf_rg_, 2);
                 /* copy to FFT buffer */
                 switch (fft_coarse_.pu()) {
-                    case CPU: {
+                case device_t::CPU: {
                         fft_coarse_.input(buf_rg_.at(memory_t::host));
                         break;
                     }
-                    case GPU: {
+                case device_t::GPU: {
 #ifdef __GPU
                         acc::copy(fft_coarse_.buffer().at(memory_t::device), buf_rg_.at(memory_t::device), fft_coarse_.local_size());
 #endif
@@ -838,7 +841,7 @@ class Local_operator
         for (int j = 0; j < phi__.pw_coeffs(0).spl_num_col().local_size(); j++) {
             utils::timer t1("sirius::Local_operator::apply_h_o|pot");
             switch (fft_coarse_.pu()) {
-                case CPU: {
+                case device_t::CPU: {
                     /* phi(G) -> phi(r) */
                     fft_coarse_.transform<1>(phi__.pw_coeffs(0).extra().at(memory_t::host, 0, j));
 
@@ -862,7 +865,7 @@ class Local_operator
                     if (hphi__ != nullptr) {
                         #pragma omp parallel for schedule(static)
                         for (int ir = 0; ir < fft_coarse_.local_size(); ir++) {
-                            /* multiply be effective potential, which itself was multiplied by the step function 
+                            /* multiply be effective potential, which itself was multiplied by the step function
                                    in the prepare() method */
                             fft_coarse_.buffer(ir) *= veff_vec_[0].f_rg(ir);
                         }
@@ -871,7 +874,7 @@ class Local_operator
                     }
                     break;
                 }
-                case GPU: {
+                case device_t::GPU: {
 #if defined(__GPU)
                     /* phi(G) -> phi(r) */
                     fft_coarse_.transform<1>(phi__.pw_coeffs(0).extra().at(memory_t::host, 0, j));
@@ -882,7 +885,8 @@ class Local_operator
                             acc::copy(buf_rg_.at(memory_t::device), fft_coarse_.buffer().at(memory_t::device), fft_coarse_.local_size());
                         }
                         /* multiply by step function */
-                        scale_matrix_rows_gpu(fft_coarse_.local_size(), 1, fft_coarse_.buffer().at(memory_t::device),
+                        scale_matrix_rows_gpu(fft_coarse_.local_size(), 1,
+                                              (acc_complex_double_t*)fft_coarse_.buffer().at(memory_t::device),
                                               theta_.f_rg().at(memory_t::device));
                         /* phi(r) * Theta(r) -> ophi(G) */
                         fft_coarse_.transform<-1>(ophi__->pw_coeffs(0).extra().at(memory_t::host, 0, j));
@@ -893,7 +897,8 @@ class Local_operator
                     }
                     if (hphi__ != nullptr) {
                         /* multiply by effective potential */
-                        scale_matrix_rows_gpu(fft_coarse_.local_size(), 1, fft_coarse_.buffer().at(memory_t::device),
+                        scale_matrix_rows_gpu(fft_coarse_.local_size(), 1,
+                                              (acc_complex_double_t*)fft_coarse_.buffer().at(memory_t::device),
                                               veff_vec_[0].f_rg().at(memory_t::device));
                         /* phi(r) * Theta(r) * V(r) -> hphi(G) */
                         fft_coarse_.transform<-1>(hphi__->pw_coeffs(0).extra().at(memory_t::host, 0, j));
@@ -918,7 +923,7 @@ class Local_operator
                     /* transform Cartesian component of wave-function gradient to real space */
                     fft_coarse_.transform<1>(&buf_pw[0]);
                     switch (fft_coarse_.pu()) {
-                        case CPU: {
+                       case device_t::CPU: {
                             #pragma omp parallel for schedule(static)
                             for (int ir = 0; ir < fft_coarse_.local_size(); ir++) {
                                 /* multiply be step function */
@@ -926,10 +931,12 @@ class Local_operator
                             }
                             break;
                         }
-                        case GPU: {
+                        case device_t::GPU: {
 #if defined(__GPU)
                             /* multiply by step function */
-                            scale_matrix_rows_gpu(fft_coarse_.local_size(), 1, fft_coarse_.buffer().at(memory_t::device), theta_.f_rg().at(memory_t::device));
+                            scale_matrix_rows_gpu(fft_coarse_.local_size(), 1,
+                                                  (acc_complex_double_t*)fft_coarse_.buffer().at(memory_t::device),
+                                                  theta_.f_rg().at(memory_t::device));
 #endif
                             break;
                         }
@@ -954,7 +961,7 @@ class Local_operator
 
         fft_coarse_.dismiss();
 
-        if (ctx_.processing_unit() == GPU) {
+        if (ctx_.processing_unit() == device_t::GPU) {
             if (hphi__ != nullptr) {
                 hphi__->pw_coeffs(0).copy_to(memory_t::device, N__, n__);
             }
@@ -1004,7 +1011,7 @@ class Local_operator
 
         for (int j = 0; j < phi__.pw_coeffs(0).spl_num_col().local_size(); j++) {
             switch (fft_coarse_.pu()) {
-                case CPU: {
+               case device_t::CPU: {
                     /* phi(G) -> phi(r) */
                     fft_coarse_.transform<1>(phi__.pw_coeffs(0).extra().at(memory_t::host, 0, j));
                     /* save phi(r) */
@@ -1030,12 +1037,14 @@ class Local_operator
                     }
                     break;
                 }
-                case GPU: {
+                case device_t::GPU: {
 #if defined(__GPU)
                     /* phi(G) -> phi(r) */
                     fft_coarse_.transform<1>(phi__.pw_coeffs(0).extra().at(memory_t::host, 0, j));
                     /* multiply by Bz */
-                    scale_matrix_rows_gpu(fft_coarse_.local_size(), 1, fft_coarse_.buffer().at(memory_t::device), veff_vec_[1].f_rg().at(memory_t::device));
+                    scale_matrix_rows_gpu(fft_coarse_.local_size(), 1,
+                                          (acc_complex_double_t*)fft_coarse_.buffer().at(memory_t::device),
+                                          veff_vec_[1].f_rg().at(memory_t::device));
                     /* phi(r) * Bz(r) -> bphi[0](G) */
                     fft_coarse_.transform<-1>(bphi__[0].pw_coeffs(0).extra().at(memory_t::host, 0, j));
 #else
