@@ -300,10 +300,10 @@ class K_point
                 //                qinv(xi2, xi1) = ctx_.augmentation_op(iat).q_mtrx(xi2, xi1);
                 //            }
                 //        }
-                //        linalg<CPU>::geinv(nbf, qinv);
+                //        linalg<device_t::CPU>::geinv(nbf, qinv);
                 //
                 //        /* compute P^{+}*P */
-                //        linalg<CPU>::gemm(2, 0, nbf, nbf, num_gkvec_loc(),
+                //        linalg<device_t::CPU>::gemm(2, 0, nbf, nbf, num_gkvec_loc(),
                 //                          beta_projectors_->beta_gk_t().at<CPU>(0, ofs), beta_projectors_->beta_gk_t().ld(),
                 //                          beta_projectors_->beta_gk_t().at<CPU>(0, ofs), beta_projectors_->beta_gk_t().ld(),
                 //                          &p_mtrx_(0, 0, iat), p_mtrx_.ld());
@@ -315,7 +315,7 @@ class K_point
                 //            }
                 //        }
                 //        /* compute (Q^{-1} + P^{+}*P)^{-1} */
-                //        linalg<CPU>::geinv(nbf, qinv);
+                //        linalg<device_t::CPU>::geinv(nbf, qinv);
                 //        for (int xi1 = 0; xi1 < nbf; xi1++) {
                 //            for (int xi2 = 0; xi2 < nbf; xi2++) {
                 //                p_mtrx_(xi2, xi1, iat) = qinv(xi2, xi1);
@@ -361,9 +361,7 @@ class K_point
         /// Generate two-component spinor wave functions
         inline void generate_spinor_wave_functions();
 
-        inline void generate_atomic_wave_functions(const int num_ao__, Wave_functions &phi);
-
-        inline void generate_atomic_wave_functions_aux(const int num_ao__, Wave_functions &phi, std::vector<int> &offset, bool hubbard);
+        inline void generate_atomic_wave_functions(const sirius::basis_functions_index&, int, int, bool, sddk::Wave_functions&);
 
         void compute_gradient_wave_functions(Wave_functions &phi,
                                              const int starting_position_i,
@@ -381,9 +379,63 @@ class K_point
 
         //== void load_wave_functions(int id);
 
-        void get_fv_eigen_vectors(mdarray<double_complex, 2>& fv_evec);
+        /// Collect distributed first-variational vectors into a global array.
+        void get_fv_eigen_vectors(mdarray<double_complex, 2>& fv_evec__) const
+        {
+            assert((int)fv_evec__.size(0) >= gklo_basis_size());
+            assert((int)fv_evec__.size(1) == ctx_.num_fv_states());
+            assert(gklo_basis_size_row() == fv_eigen_vectors_.num_rows_local());
 
-        void get_sv_eigen_vectors(mdarray<double_complex, 2>& sv_evec);
+            mdarray<double_complex, 1> tmp(gklo_basis_size_row());
+
+            fv_evec__.zero();
+
+            for (int ist = 0; ist < ctx_.num_fv_states(); ist++) {
+                auto loc = fv_eigen_vectors_.spl_col().location(ist);
+                if (loc.rank == fv_eigen_vectors_.rank_col()) {
+                    std::copy(&fv_eigen_vectors_(0, loc.local_index),
+                              &fv_eigen_vectors_(0, loc.local_index) + gklo_basis_size_row(),
+                              &tmp(0));
+                }
+                fv_eigen_vectors_.blacs_grid().comm_col().bcast(&tmp(0), gklo_basis_size_row(), loc.rank);
+                for (int jloc = 0; jloc < gklo_basis_size_row(); jloc++) {
+                    int j = fv_eigen_vectors_.irow(jloc);
+                    fv_evec__(j, ist) = tmp(jloc);
+                }
+                fv_eigen_vectors_.blacs_grid().comm_row().allreduce(&fv_evec__(0, ist), gklo_basis_size());
+            }
+        }
+
+        /// Collect distributed second-variational vectors into a global array.
+        void get_sv_eigen_vectors(mdarray<double_complex, 2>& sv_evec) const
+        {
+            assert((int)sv_evec.size(0) == ctx_.num_bands());
+            assert((int)sv_evec.size(1) == ctx_.num_bands());
+
+            sv_evec.zero();
+
+            if (!ctx_.need_sv()) {
+                for (int i = 0; i < ctx_.num_fv_states(); i++) {
+                    sv_evec(i, i) = 1;
+                }
+                return;
+            }
+
+            int nsp = (ctx_.num_mag_dims() == 3) ? 1 : ctx_.num_spins();
+
+            for (int ispn = 0; ispn < nsp; ispn++) {
+                int offs = ctx_.num_fv_states() * ispn;
+                for (int jloc = 0; jloc < sv_eigen_vectors_[ispn].num_cols_local(); jloc++) {
+                    int j = sv_eigen_vectors_[ispn].icol(jloc);
+                    for (int iloc = 0; iloc < sv_eigen_vectors_[ispn].num_rows_local(); iloc++) {
+                        int i = sv_eigen_vectors_[ispn].irow(iloc);
+                        sv_evec(i + offs, j + offs) = sv_eigen_vectors_[ispn](iloc, jloc);
+                    }
+                }
+            }
+
+            comm_.allreduce(sv_evec.at(memory_t::host), (int)sv_evec.size());
+        }
 
         /// Test orthonormalization of spinor wave-functions
         void test_spinor_wave_functions(int use_fft);
@@ -515,9 +567,11 @@ class K_point
         }
 
         /// Basis size of LAPW+lo method.
+        /** The total LAPW+lo basis size is equal to the sum of the number of LAPW functions and the total number
+         *  of the local orbitals. */
         inline int gklo_basis_size() const
         {
-            return static_cast<int>(num_gkvec() + unit_cell_.mt_lo_basis_size());
+            return num_gkvec() + unit_cell_.mt_lo_basis_size();
         }
 
         /// Local number of G+k vectors in case of flat distribution.
@@ -746,6 +800,12 @@ class K_point
             return *beta_projectors_;
         }
 
+        Beta_projectors const& beta_projectors() const
+        {
+            assert(beta_projectors_ != nullptr);
+            return *beta_projectors_;
+        }
+
         Beta_projectors& beta_projectors_row()
         {
             assert(beta_projectors_ != nullptr);
@@ -762,16 +822,28 @@ class K_point
         {
             return ctx_;
         }
+
+        /// Print message from the root rank.
+        template <typename... Args>
+        inline void message(int level__, char const* label__, Args... args) const
+        {
+            if (this->comm().rank() == 0 && this->ctx().control().verbosity_ >= level__) {
+                if (label__) {
+                    printf("[%s] ", label__);
+                }
+                printf(args...);
+            }
+        }
 };
 
 //== void K_point::check_alm(int num_gkvec_loc, int ia, mdarray<double_complex, 2>& alm)
 //== {
 //==     static SHT* sht = NULL;
 //==     if (!sht) sht = new SHT(ctx_.lmax_apw());
-//== 
+//==
 //==     Atom* atom = unit_cell_.atom(ia);
 //==     Atom_type* type = atom->type();
-//== 
+//==
 //==     mdarray<double_complex, 2> z1(sht->num_points(), type->mt_aw_basis_size());
 //==     for (int i = 0; i < type->mt_aw_basis_size(); i++)
 //==     {
@@ -783,13 +855,13 @@ class K_point
 //==             z1(itp, i) = sht->ylm_backward(lm, itp) * rf;
 //==         }
 //==     }
-//== 
+//==
 //==     mdarray<double_complex, 2> z2(sht->num_points(), num_gkvec_loc);
 //==     blas<CPU>::gemm(0, 2, sht->num_points(), num_gkvec_loc, type->mt_aw_basis_size(), z1.ptr(), z1.ld(),
 //==                     alm.ptr(), alm.ld(), z2.ptr(), z2.ld());
-//== 
+//==
 //==     vector3d<double> vc = unit_cell_.get_cartesian_coordinates(unit_cell_.atom(ia)->position());
-//==     
+//==
 //==     double tdiff = 0;
 //==     for (int igloc = 0; igloc < num_gkvec_loc; igloc++)
 //==     {
@@ -803,8 +875,8 @@ class K_point
 //==             tdiff += abs(pw_value - aw_value);
 //==         }
 //==     }
-//== 
-//==     printf("atom : %i  absolute alm error : %e  average alm error : %e\n", 
+//==
+//==     printf("atom : %i  absolute alm error : %e  average alm error : %e\n",
 //==            ia, tdiff, tdiff / (num_gkvec_loc * sht->num_points()));
 //== }
 
@@ -815,17 +887,17 @@ class K_point
 //
 //    int lmmax = Utils::lmmax_by_lmax(lmax);
 //
-//    Periodic_function<double_complex, index_order>* func = 
+//    Periodic_function<double_complex, index_order>* func =
 //        new Periodic_function<double_complex, index_order>(ctx_, lmax);
 //    func->allocate(ylm_component | it_component);
 //    func->zero();
-//    
+//
 //    if (basis_type == pwlo)
 //    {
 //        if (index_order != radial_angular) error(__FILE__, __LINE__, "wrong order of indices");
 //
 //        double fourpi_omega = fourpi / sqrt(ctx_.omega());
-//        
+//
 //        for (int igkloc = 0; igkloc < num_gkvec_row(); igkloc++)
 //        {
 //            int igk = igkglob(igkloc);
@@ -836,12 +908,12 @@ class K_point
 //            {
 //                int iat = ctx_.atom_type_index_by_id(ctx_.atom(ia)->type_id());
 //                double_complex z2 = z1 * gkvec_phase_factors_(igkloc, ia);
-//                
+//
 //                #pragma omp parallel for default(shared)
 //                for (int lm = 0; lm < lmmax; lm++)
 //                {
 //                    int l = l_by_lm_(lm);
-//                    double_complex z3 = z2 * zil_[l] * conj(gkvec_ylm_(lm, igkloc)); 
+//                    double_complex z3 = z2 * zil_[l] * conj(gkvec_ylm_(lm, igkloc));
 //                    for (int ir = 0; ir < ctx_.atom(ia)->num_mt_points(); ir++)
 //                        func->f_ylm(ir, lm, ia) += z3 * (*sbessel_[igkloc])(ir, l, iat);
 //                }
@@ -867,8 +939,8 @@ class K_point
 //                {
 //                    for (int ir = 0; ir < ctx_.atom(ia)->num_mt_points(); ir++)
 //                    {
-//                        func->f_ylm(lm, ir, ia) += 
-//                            spinor_wave_functions_(ctx_.atom(ia)->offset_wf() + i, ispn, jloc) * 
+//                        func->f_ylm(lm, ir, ia) +=
+//                            spinor_wave_functions_(ctx_.atom(ia)->offset_wf() + i, ispn, jloc) *
 //                            ctx_.atom(ia)->symmetry_class()->radial_function(ir, idxrf);
 //                    }
 //                    break;
@@ -877,8 +949,8 @@ class K_point
 //                {
 //                    for (int ir = 0; ir < ctx_.atom(ia)->num_mt_points(); ir++)
 //                    {
-//                        func->f_ylm(ir, lm, ia) += 
-//                            spinor_wave_functions_(ctx_.atom(ia)->offset_wf() + i, ispn, jloc) * 
+//                        func->f_ylm(ir, lm, ia) +=
+//                            spinor_wave_functions_(ctx_.atom(ia)->offset_wf() + i, ispn, jloc) *
 //                            ctx_.atom(ia)->symmetry_class()->radial_function(ir, idxrf);
 //                    }
 //                    break;
@@ -888,32 +960,32 @@ class K_point
 //    }
 //
 //    // in principle, wave function must have an overall e^{ikr} phase factor
-//    ctx_.fft().input(num_gkvec(), &fft_index_[0], 
+//    ctx_.fft().input(num_gkvec(), &fft_index_[0],
 //                            &spinor_wave_functions_(ctx_.mt_basis_size(), ispn, jloc));
 //    ctx_.fft().transform(1);
 //    ctx_.fft().output(func->f_it());
 //
 //    for (int i = 0; i < ctx_.fft().size(); i++) func->f_it(i) /= sqrt(ctx_.omega());
-//    
+//
 //    return func;
 //}
 
 //== void K_point::spinor_wave_function_component_mt(int lmax, int ispn, int jloc, mt_functions<double_complex>& psilm)
 //== {
 //==     Timer t("sirius::K_point::spinor_wave_function_component_mt");
-//== 
+//==
 //==     //int lmmax = Utils::lmmax_by_lmax(lmax);
-//== 
+//==
 //==     psilm.zero();
-//==     
+//==
 //==     //if (basis_type == pwlo)
 //==     //{
 //==     //    if (index_order != radial_angular) error(__FILE__, __LINE__, "wrong order of indices");
-//== 
+//==
 //==     //    double fourpi_omega = fourpi / sqrt(ctx_.omega());
-//== 
+//==
 //==     //    mdarray<double_complex, 2> zm(ctx_.max_num_mt_points(),  num_gkvec_row());
-//== 
+//==
 //==     //    for (int ia = 0; ia < ctx_.num_atoms(); ia++)
 //==     //    {
 //==     //        int iat = ctx_.atom_type_index_by_id(ctx_.atom(ia)->type_id());
@@ -929,7 +1001,7 @@ class K_point
 //==     //                    zm(ir, igkloc) = z2 * (*sbessel_[igkloc])(ir, l, iat);
 //==     //            }
 //==     //            blas<CPU>::gemm(0, 2, ctx_.atom(ia)->num_mt_points(), (2 * l + 1), num_gkvec_row(),
-//==     //                            &zm(0, 0), zm.ld(), &gkvec_ylm_(Utils::lm_by_l_m(l, -l), 0), gkvec_ylm_.ld(), 
+//==     //                            &zm(0, 0), zm.ld(), &gkvec_ylm_(Utils::lm_by_l_m(l, -l), 0), gkvec_ylm_.ld(),
 //==     //                            &fylm(0, Utils::lm_by_l_m(l, -l), ia), fylm.ld());
 //==     //        }
 //==     //    }
@@ -937,31 +1009,31 @@ class K_point
 //==     //    //{
 //==     //    //    int igk = igkglob(igkloc);
 //==     //    //    double_complex z1 = spinor_wave_functions_(ctx_.mt_basis_size() + igk, ispn, jloc) * fourpi_omega;
-//== 
+//==
 //==     //    //    // TODO: possilbe optimization with zgemm
 //==     //    //    for (int ia = 0; ia < ctx_.num_atoms(); ia++)
 //==     //    //    {
 //==     //    //        int iat = ctx_.atom_type_index_by_id(ctx_.atom(ia)->type_id());
 //==     //    //        double_complex z2 = z1 * gkvec_phase_factors_(igkloc, ia);
-//==     //    //        
+//==     //    //
 //==     //    //        #pragma omp parallel for default(shared)
 //==     //    //        for (int lm = 0; lm < lmmax; lm++)
 //==     //    //        {
 //==     //    //            int l = l_by_lm_(lm);
-//==     //    //            double_complex z3 = z2 * zil_[l] * conj(gkvec_ylm_(lm, igkloc)); 
+//==     //    //            double_complex z3 = z2 * zil_[l] * conj(gkvec_ylm_(lm, igkloc));
 //==     //    //            for (int ir = 0; ir < ctx_.atom(ia)->num_mt_points(); ir++)
 //==     //    //                fylm(ir, lm, ia) += z3 * (*sbessel_[igkloc])(ir, l, iat);
 //==     //    //        }
 //==     //    //    }
 //==     //    //}
-//== 
+//==
 //==     //    for (int ia = 0; ia < ctx_.num_atoms(); ia++)
 //==     //    {
 //==     //        Platform::allreduce(&fylm(0, 0, ia), lmmax * ctx_.max_num_mt_points(),
 //==     //                            ctx_.mpi_grid().communicator(1 << band->dim_row()));
 //==     //    }
 //==     //}
-//== 
+//==
 //==     for (int ia = 0; ia < ctx_.num_atoms(); ia++)
 //==     {
 //==         for (int i = 0; i < ctx_.atom(ia)->type()->mt_basis_size(); i++)
@@ -970,8 +1042,8 @@ class K_point
 //==             int idxrf = ctx_.atom(ia)->type()->indexb(i).idxrf;
 //==             for (int ir = 0; ir < ctx_.atom(ia)->num_mt_points(); ir++)
 //==             {
-//==                 psilm(lm, ir, ia) += 
-//==                     spinor_wave_functions_(ctx_.atom(ia)->offset_wf() + i, ispn, jloc) * 
+//==                 psilm(lm, ir, ia) +=
+//==                     spinor_wave_functions_(ctx_.atom(ia)->offset_wf() + i, ispn, jloc) *
 //==                     ctx_.atom(ia)->symmetry_class()->radial_function(ir, idxrf);
 //==             }
 //==         }
@@ -983,24 +1055,24 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
     STOP();
 
 //==     if (num_ranks() > 1) error_local(__FILE__, __LINE__, "test of spinor wave functions on multiple ranks is not implemented");
-//== 
+//==
 //==     std::vector<double_complex> v1[2];
 //==     std::vector<double_complex> v2;
-//== 
+//==
 //==     if (use_fft == 0 || use_fft == 1) v2.resize(fft_->size());
-//==     
-//==     if (use_fft == 0) 
+//==
+//==     if (use_fft == 0)
 //==     {
 //==         for (int ispn = 0; ispn < ctx_.num_spins(); ispn++) v1[ispn].resize(num_gkvec());
 //==     }
-//==     
-//==     if (use_fft == 1) 
+//==
+//==     if (use_fft == 1)
 //==     {
 //==         for (int ispn = 0; ispn < ctx_.num_spins(); ispn++) v1[ispn].resize(fft_->size());
 //==     }
-//==     
+//==
 //==     double maxerr = 0;
-//== 
+//==
 //==     for (int j1 = 0; j1 < ctx_.num_bands(); j1++)
 //==     {
 //==         if (use_fft == 0)
@@ -1011,15 +1083,15 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
 //==                                        &spinor_wave_functions_(unit_cell_.mt_basis_size(), ispn, j1));
 //==                 fft_->transform(1);
 //==                 fft_->output(&v2[0]);
-//== 
+//==
 //==                 for (int ir = 0; ir < fft_->size(); ir++) v2[ir] *= ctx_.step_function()->theta_r(ir);
-//==                 
+//==
 //==                 fft_->input(&v2[0]);
 //==                 fft_->transform(-1);
-//==                 fft_->output(num_gkvec(), gkvec_.index_map(), &v1[ispn][0]); 
+//==                 fft_->output(num_gkvec(), gkvec_.index_map(), &v1[ispn][0]);
 //==             }
 //==         }
-//==         
+//==
 //==         if (use_fft == 1)
 //==         {
 //==             for (int ispn = 0; ispn < ctx_.num_spins(); ispn++)
@@ -1030,7 +1102,7 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
 //==                 fft_->output(&v1[ispn][0]);
 //==             }
 //==         }
-//==        
+//==
 //==         for (int j2 = 0; j2 < ctx_.num_bands(); j2++)
 //==         {
 //==             double_complex zsum(0, 0);
@@ -1041,7 +1113,7 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
 //==                     int offset_wf = unit_cell_.atom(ia)->offset_wf();
 //==                     Atom_type* type = unit_cell_.atom(ia)->type();
 //==                     Atom_symmetry_class* symmetry_class = unit_cell_.atom(ia)->symmetry_class();
-//== 
+//==
 //==                     for (int l = 0; l <= ctx_.lmax_apw(); l++)
 //==                     {
 //==                         int ordmax = type->indexr().num_rf(l);
@@ -1052,7 +1124,7 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
 //==                                 for (int m = -l; m <= l; m++)
 //==                                 {
 //==                                     zsum += conj(spinor_wave_functions_(offset_wf + type->indexb_by_l_m_order(l, m, io1), ispn, j1)) *
-//==                                             spinor_wave_functions_(offset_wf + type->indexb_by_l_m_order(l, m, io2), ispn, j2) * 
+//==                                             spinor_wave_functions_(offset_wf + type->indexb_by_l_m_order(l, m, io2), ispn, j2) *
 //==                                             symmetry_class->o_radial_integral(l, io1, io2);
 //==                                 }
 //==                             }
@@ -1060,7 +1132,7 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
 //==                     }
 //==                 }
 //==             }
-//==             
+//==
 //==             if (use_fft == 0)
 //==             {
 //==                for (int ispn = 0; ispn < ctx_.num_spins(); ispn++)
@@ -1069,7 +1141,7 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
 //==                        zsum += conj(v1[ispn][ig]) * spinor_wave_functions_(unit_cell_.mt_basis_size() + ig, ispn, j2);
 //==                }
 //==             }
-//==            
+//==
 //==             if (use_fft == 1)
 //==             {
 //==                 for (int ispn = 0; ispn < ctx_.num_spins(); ispn++)
@@ -1077,13 +1149,13 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
 //==                     fft_->input(num_gkvec(), gkvec_.index_map(), &spinor_wave_functions_(unit_cell_.mt_basis_size(), ispn, j2));
 //==                     fft_->transform(1);
 //==                     fft_->output(&v2[0]);
-//== 
+//==
 //==                     for (int ir = 0; ir < fft_->size(); ir++)
 //==                         zsum += std::conj(v1[ispn][ir]) * v2[ir] * ctx_.step_function()->theta_r(ir) / double(fft_->size());
 //==                 }
 //==             }
-//==             
-//==             if (use_fft == 2) 
+//==
+//==             if (use_fft == 2)
 //==             {
 //==                 STOP();
 //==                 //for (int ig1 = 0; ig1 < num_gkvec(); ig1++)
@@ -1093,14 +1165,14 @@ inline void K_point::test_spinor_wave_functions(int use_fft)
 //==                 //        int ig3 = ctx_.gvec().index_g12(ig1, ig2);
 //==                 //        for (int ispn = 0; ispn < ctx_.num_spins(); ispn++)
 //==                 //        {
-//==                 //            zsum += std::conj(spinor_wave_functions_(unit_cell_.mt_basis_size() + ig1, ispn, j1)) * 
-//==                 //                    spinor_wave_functions_(unit_cell_.mt_basis_size() + ig2, ispn, j2) * 
+//==                 //            zsum += std::conj(spinor_wave_functions_(unit_cell_.mt_basis_size() + ig1, ispn, j1)) *
+//==                 //                    spinor_wave_functions_(unit_cell_.mt_basis_size() + ig2, ispn, j2) *
 //==                 //                    ctx_.step_function()->theta_pw(ig3);
 //==                 //        }
 //==                 //    }
 //==                 //}
 //==             }
-//== 
+//==
 //==             zsum = (j1 == j2) ? zsum - double_complex(1.0, 0.0) : zsum;
 //==             maxerr = std::max(maxerr, std::abs(zsum));
 //==         }
@@ -1188,7 +1260,7 @@ inline void K_point::load(HDF5_tree h5in, int id)
 
     //== band_occupancies_.resize(ctx_.num_bands());
     //== h5in[id].read("band_occupancies", band_occupancies_);
-    //== 
+    //==
     //== h5in[id].read_mdarray("fv_eigen_vectors", fv_eigen_vectors_panel_);
     //== h5in[id].read_mdarray("sv_eigen_vectors", sv_eigen_vectors_);
 }
@@ -1198,7 +1270,7 @@ inline void K_point::load(HDF5_tree h5in, int id)
 //==     if (ctx_.mpi_grid().root(1 << _dim_col_))
 //==     {
 //==         HDF5_tree fout(storage_file_name, false);
-//== 
+//==
 //==         fout["K_points"].create_node(id);
 //==         fout["K_points"][id].write("coordinates", &vk_[0], 3);
 //==         fout["K_points"][id].write("mtgk_size", mtgk_size());
@@ -1206,10 +1278,10 @@ inline void K_point::load(HDF5_tree h5in, int id)
 //==         fout["K_points"][id].write("band_energies", &band_energies_[0], ctx_.num_bands());
 //==         fout["K_points"][id].write("band_occupancies", &band_occupancies_[0], ctx_.num_bands());
 //==     }
-//==     
+//==
 //==     Platform::barrier(ctx_.mpi_grid().communicator(1 << _dim_col_));
-//==     
-//==     mdarray<double_complex, 2> wfj(NULL, mtgk_size(), ctx_.num_spins()); 
+//==
+//==     mdarray<double_complex, 2> wfj(NULL, mtgk_size(), ctx_.num_spins());
 //==     for (int j = 0; j < ctx_.num_bands(); j++)
 //==     {
 //==         int rank = ctx_.spl_spinor_wf_col().location(_splindex_rank_, j);
@@ -1223,26 +1295,26 @@ inline void K_point::load(HDF5_tree h5in, int id)
 //==         Platform::barrier(ctx_.mpi_grid().communicator(_dim_col_));
 //==     }
 //== }
-//== 
+//==
 //== void K_point::load_wave_functions(int id)
 //== {
 //==     HDF5_tree fin(storage_file_name, false);
-//==     
+//==
 //==     int mtgk_size_in;
 //==     fin["K_points"][id].read("mtgk_size", &mtgk_size_in);
 //==     if (mtgk_size_in != mtgk_size()) error_local(__FILE__, __LINE__, "wrong wave-function size");
-//== 
+//==
 //==     band_energies_.resize(ctx_.num_bands());
 //==     fin["K_points"][id].read("band_energies", &band_energies_[0], ctx_.num_bands());
-//== 
+//==
 //==     band_occupancies_.resize(ctx_.num_bands());
 //==     fin["K_points"][id].read("band_occupancies", &band_occupancies_[0], ctx_.num_bands());
-//== 
-//==     spinor_wave_functions_.set_dimensions(mtgk_size(), ctx_.num_spins(), 
+//==
+//==     spinor_wave_functions_.set_dimensions(mtgk_size(), ctx_.num_spins(),
 //==                                           ctx_.spl_spinor_wf_col().local_size());
 //==     spinor_wave_functions_.allocate();
-//== 
-//==     mdarray<double_complex, 2> wfj(NULL, mtgk_size(), ctx_.num_spins()); 
+//==
+//==     mdarray<double_complex, 2> wfj(NULL, mtgk_size(), ctx_.num_spins());
 //==     for (int jloc = 0; jloc < ctx_.spl_spinor_wf_col().local_size(); jloc++)
 //==     {
 //==         int j = ctx_.spl_spinor_wf_col(jloc);
@@ -1251,59 +1323,6 @@ inline void K_point::load(HDF5_tree h5in, int id)
 //==     }
 //== }
 
-inline void K_point::get_fv_eigen_vectors(mdarray<double_complex, 2>& fv_evec)
-{
-    assert((int)fv_evec.size(0) >= gklo_basis_size());
-    assert((int)fv_evec.size(1) == ctx_.num_fv_states());
-    
-    fv_evec.zero();
-    STOP();
-
-    //for (int iloc = 0; iloc < (int)spl_fv_states_.local_size(); iloc++)
-    //{
-    //    int i = (int)spl_fv_states_[iloc];
-    //    for (int jloc = 0; jloc < gklo_basis_size_row(); jloc++)
-    //    {
-    //        int j = gklo_basis_descriptor_row(jloc).id;
-    //        fv_evec(j, i) = fv_eigen_vectors_(jloc, iloc);
-    //    }
-    //}
-    //comm_.allreduce(fv_evec.at(memory_t::host), (int)fv_evec.size());
-}
-
-inline void K_point::get_sv_eigen_vectors(mdarray<double_complex, 2>& sv_evec)
-{
-    assert((int)sv_evec.size(0) == ctx_.num_bands());
-    assert((int)sv_evec.size(1) == ctx_.num_bands());
-
-    sv_evec.zero();
-
-    if (!ctx_.need_sv()) {
-        for (int i = 0; i < ctx_.num_fv_states(); i++) {
-            sv_evec(i, i) = 1;
-        }
-        return;
-    }
-
-    int nsp = (ctx_.num_mag_dims() == 3) ? 1 : ctx_.num_spins();
-
-    for (int ispn = 0; ispn < nsp; ispn++)
-    {
-        int offs = ctx_.num_fv_states() * ispn;
-        for (int jloc = 0; jloc < sv_eigen_vectors_[ispn].num_cols_local(); jloc++)
-        {
-            int j = sv_eigen_vectors_[ispn].icol(jloc);
-            for (int iloc = 0; iloc < sv_eigen_vectors_[ispn].num_rows_local(); iloc++)
-            {
-                int i = sv_eigen_vectors_[ispn].irow(iloc);
-                sv_evec(i + offs, j + offs) = sv_eigen_vectors_[ispn](iloc, jloc);
-            }
-        }
-    }
-
-    comm_.allreduce(sv_evec.at(memory_t::host), (int)sv_evec.size());
-}
-
 #include "initialize.hpp"
 #include "generate_fv_states.hpp"
 #include "generate_spinor_wave_functions.hpp"
