@@ -24,18 +24,21 @@
 
 inline void Potential::xc_mt_nonmagnetic(Radial_grid<double> const& rgrid,
                                          std::vector<XC_functional>& xc_func,
-                                         Spheric_function<spectral, double> const& rho_lm,
-                                         Spheric_function<spatial, double>& rho_tp,
-                                         Spheric_function<spatial, double>& vxc_tp,
-                                         Spheric_function<spatial, double>& exc_tp)
+                                         Spheric_function<function_domain_t::spectral, double> const& rho_lm,
+                                         Spheric_function<function_domain_t::spatial, double>& rho_tp,
+                                         Spheric_function<function_domain_t::spatial, double>& vxc_tp,
+                                         Spheric_function<function_domain_t::spatial, double>& exc_tp)
 {
     PROFILE("sirius::Potential::xc_mt_nonmagnetic");
 
+    /* use Laplacian (true) or divergence of gradient (false) */
+    bool use_lapl{false};
+
     bool is_gga = is_gradient_correction();
 
-    Spheric_function_gradient<spatial, double> grad_rho_tp(sht_->num_points(), rgrid);
-    Spheric_function<spatial, double> lapl_rho_tp;
-    Spheric_function<spatial, double> grad_rho_grad_rho_tp;
+    Spheric_vector_function<function_domain_t::spatial, double> grad_rho_tp(sht_->num_points(), rgrid);
+    Spheric_function<function_domain_t::spatial, double> lapl_rho_tp;
+    Spheric_function<function_domain_t::spatial, double> grad_rho_grad_rho_tp;
 
     if (is_gga) {
         /* compute gradient in Rlm spherical harmonics */
@@ -43,25 +46,28 @@ inline void Potential::xc_mt_nonmagnetic(Radial_grid<double> const& rgrid,
 
         /* backward transform gradient from Rlm to (theta, phi) */
         for (int x = 0; x < 3; x++) {
-            grad_rho_tp[x] = transform(sht_.get(), grad_rho_lm[x]);
+            transform(*sht_, grad_rho_lm[x], grad_rho_tp[x]);
         }
 
         /* compute density gradient product */
         grad_rho_grad_rho_tp = grad_rho_tp * grad_rho_tp;
 
-        /* compute Laplacian in Rlm spherical harmonics */
-        auto lapl_rho_lm = laplacian(rho_lm);
+        if (use_lapl) {
+            lapl_rho_tp = Spheric_function<function_domain_t::spatial, double>(sht_->num_points(), rgrid);
+            /* compute Laplacian in Rlm spherical harmonics */
+            auto lapl_rho_lm = laplacian(rho_lm);
 
-        /* backward transform Laplacian from Rlm to (theta, phi) */
-        lapl_rho_tp = transform(sht_.get(), lapl_rho_lm);
+            /* backward transform Laplacian from Rlm to (theta, phi) */
+            transform(*sht_, lapl_rho_lm, lapl_rho_tp);
+        }
     }
 
     exc_tp.zero();
     vxc_tp.zero();
 
-    Spheric_function<spatial, double> vsigma_tp;
+    Spheric_function<function_domain_t::spatial, double> vsigma_tp;
     if (is_gga) {
-        vsigma_tp = Spheric_function<spatial, double>(sht_->num_points(), rgrid);
+        vsigma_tp = Spheric_function<function_domain_t::spatial, double>(sht_->num_points(), rgrid);
         vsigma_tp.zero();
     }
 
@@ -71,7 +77,6 @@ inline void Potential::xc_mt_nonmagnetic(Radial_grid<double> const& rgrid,
         if (ixc.is_lda()) {
             #pragma omp parallel
             {
-
                 std::vector<double> exc_t(sht_->num_points());
                 std::vector<double> vxc_t(sht_->num_points());
                 #pragma omp for
@@ -94,16 +99,18 @@ inline void Potential::xc_mt_nonmagnetic(Radial_grid<double> const& rgrid,
                 std::vector<double> vrho_t(sht_->num_points());
                 std::vector<double> vsigma_t(sht_->num_points());
                 #pragma omp for
-                for (int ir = 0; ir < rgrid.num_points(); ir++)
-                {
+                for (int ir = 0; ir < rgrid.num_points(); ir++) {
                     ixc.get_gga(sht_->num_points(), &rho_tp(0, ir), &grad_rho_grad_rho_tp(0, ir), &vrho_t[0], &vsigma_t[0], &exc_t[0]);
-                    for (int itp = 0; itp < sht_->num_points(); itp++)
-                    {
+                    for (int itp = 0; itp < sht_->num_points(); itp++) {
                         /* add Exc contribution */
                         exc_tp(itp, ir) += exc_t[itp];
 
                         /* directly add to Vxc available contributions */
-                        vxc_tp(itp, ir) += (vrho_t[itp] - 2 * vsigma_t[itp] * lapl_rho_tp(itp, ir));
+                        vxc_tp(itp, ir) += vrho_t[itp];
+
+                        if (use_lapl) {
+                            vxc_tp(itp, ir) -= 2 * vsigma_t[itp] * lapl_rho_tp(itp, ir);
+                        }
 
                         /* save the sigma derivative */
                         vsigma_tp(itp, ir) += vsigma_t[itp];
@@ -113,29 +120,43 @@ inline void Potential::xc_mt_nonmagnetic(Radial_grid<double> const& rgrid,
         }
     }
 
-    if (is_gga)
-    {
-        /* forward transform vsigma to Rlm */
-        auto vsigma_lm = transform(sht_.get(), vsigma_tp);
+    if (is_gga) {
+        if (use_lapl) {
+            Spheric_function<function_domain_t::spectral, double> vsigma_lm(ctx_.lmmax_pot(), rgrid);
+            /* forward transform vsigma to Rlm */
+            transform(*sht_, vsigma_tp, vsigma_lm);
 
-        /* compute gradient of vsgima in spherical harmonics */
-        auto grad_vsigma_lm = gradient(vsigma_lm);
+            /* compute gradient of vsgima in spherical harmonics */
+            auto grad_vsigma_lm = gradient(vsigma_lm);
 
-        /* backward transform gradient from Rlm to (theta, phi) */
-        Spheric_function_gradient<spatial, double> grad_vsigma_tp(sht_->num_points(), rgrid);
-        for (int x = 0; x < 3; x++) {
-            grad_vsigma_tp[x] = transform(sht_.get(), grad_vsigma_lm[x]);
-        }
+            /* backward transform gradient from Rlm to (theta, phi) */
+            Spheric_vector_function<function_domain_t::spatial, double> grad_vsigma_tp(sht_->num_points(), rgrid);
+            for (int x = 0; x < 3; x++) {
+                transform(*sht_, grad_vsigma_lm[x], grad_vsigma_tp[x]);
+            }
 
-        /* compute scalar product of two gradients */
-        auto grad_vsigma_grad_rho_tp = grad_vsigma_tp * grad_rho_tp;
+            /* compute scalar product of two gradients */
+            auto grad_vsigma_grad_rho_tp = grad_vsigma_tp * grad_rho_tp;
 
-        /* add remaining term to Vxc */
-        for (int ir = 0; ir < rgrid.num_points(); ir++)
-        {
-            for (int itp = 0; itp < sht_->num_points(); itp++)
-            {
-                vxc_tp(itp, ir) -= 2 * grad_vsigma_grad_rho_tp(itp, ir);
+            /* add remaining term to Vxc */
+            for (int ir = 0; ir < rgrid.num_points(); ir++) {
+                for (int itp = 0; itp < sht_->num_points(); itp++) {
+                    vxc_tp(itp, ir) -= 2 * grad_vsigma_grad_rho_tp(itp, ir);
+                }
+            }
+        } else {
+            Spheric_vector_function<function_domain_t::spectral, double> vsigma_grad_rho_lm(ctx_.lmmax_pot(), rgrid);
+            for (int x: {0, 1, 2}) {
+                auto vsigma_grad_rho_tp = vsigma_tp * grad_rho_tp[x];
+                transform(*sht_, vsigma_grad_rho_tp, vsigma_grad_rho_lm[x]);
+            }
+            auto div_vsigma_grad_rho_lm = divergence(vsigma_grad_rho_lm);
+            auto div_vsigma_grad_rho_tp = transform(*sht_, div_vsigma_grad_rho_lm);
+            /* add remaining term to Vxc */
+            for (int ir = 0; ir < rgrid.num_points(); ir++) {
+                for (int itp = 0; itp < sht_->num_points(); itp++) {
+                    vxc_tp(itp, ir) -= 2 * div_vsigma_grad_rho_tp(itp, ir);
+                }
             }
         }
     }
@@ -143,27 +164,27 @@ inline void Potential::xc_mt_nonmagnetic(Radial_grid<double> const& rgrid,
 
 inline void Potential::xc_mt_magnetic(Radial_grid<double> const& rgrid,
                                       std::vector<XC_functional>& xc_func,
-                                      Spheric_function<spectral, double>& rho_up_lm,
-                                      Spheric_function<spatial, double>& rho_up_tp,
-                                      Spheric_function<spectral, double>& rho_dn_lm,
-                                      Spheric_function<spatial, double>& rho_dn_tp,
-                                      Spheric_function<spatial, double>& vxc_up_tp,
-                                      Spheric_function<spatial, double>& vxc_dn_tp,
-                                      Spheric_function<spatial, double>& exc_tp)
+                                      Spheric_function<function_domain_t::spectral, double>& rho_up_lm,
+                                      Spheric_function<function_domain_t::spatial, double>& rho_up_tp,
+                                      Spheric_function<function_domain_t::spectral, double>& rho_dn_lm,
+                                      Spheric_function<function_domain_t::spatial, double>& rho_dn_tp,
+                                      Spheric_function<function_domain_t::spatial, double>& vxc_up_tp,
+                                      Spheric_function<function_domain_t::spatial, double>& vxc_dn_tp,
+                                      Spheric_function<function_domain_t::spatial, double>& exc_tp)
 {
     PROFILE("sirius::Potential::xc_mt_magnetic");
 
     bool is_gga = is_gradient_correction();
 
-    Spheric_function_gradient<spatial, double> grad_rho_up_tp(sht_->num_points(), rgrid);
-    Spheric_function_gradient<spatial, double> grad_rho_dn_tp(sht_->num_points(), rgrid);
+    Spheric_vector_function<function_domain_t::spatial, double> grad_rho_up_tp(sht_->num_points(), rgrid);
+    Spheric_vector_function<function_domain_t::spatial, double> grad_rho_dn_tp(sht_->num_points(), rgrid);
 
-    Spheric_function<spatial, double> lapl_rho_up_tp(sht_->num_points(), rgrid);
-    Spheric_function<spatial, double> lapl_rho_dn_tp(sht_->num_points(), rgrid);
+    Spheric_function<function_domain_t::spatial, double> lapl_rho_up_tp(sht_->num_points(), rgrid);
+    Spheric_function<function_domain_t::spatial, double> lapl_rho_dn_tp(sht_->num_points(), rgrid);
 
-    Spheric_function<spatial, double> grad_rho_up_grad_rho_up_tp;
-    Spheric_function<spatial, double> grad_rho_dn_grad_rho_dn_tp;
-    Spheric_function<spatial, double> grad_rho_up_grad_rho_dn_tp;
+    Spheric_function<function_domain_t::spatial, double> grad_rho_up_grad_rho_up_tp;
+    Spheric_function<function_domain_t::spatial, double> grad_rho_dn_grad_rho_dn_tp;
+    Spheric_function<function_domain_t::spatial, double> grad_rho_up_grad_rho_dn_tp;
 
     //assert(rho_up_lm.radial_grid().hash() == rho_dn_lm.radial_grid().hash());
 
@@ -171,17 +192,15 @@ inline void Potential::xc_mt_magnetic(Radial_grid<double> const& rgrid,
     vxc_dn_tp.zero();
     exc_tp.zero();
 
-    if (is_gga)
-    {
+    if (is_gga) {
         /* compute gradient in Rlm spherical harmonics */
         auto grad_rho_up_lm = gradient(rho_up_lm);
         auto grad_rho_dn_lm = gradient(rho_dn_lm);
 
         /* backward transform gradient from Rlm to (theta, phi) */
-        for (int x = 0; x < 3; x++)
-        {
-            grad_rho_up_tp[x] = transform(sht_.get(), grad_rho_up_lm[x]);
-            grad_rho_dn_tp[x] = transform(sht_.get(), grad_rho_dn_lm[x]);
+        for (int x = 0; x < 3; x++) {
+            grad_rho_up_tp[x] = transform(*sht_, grad_rho_up_lm[x]);
+            grad_rho_dn_tp[x] = transform(*sht_, grad_rho_dn_lm[x]);
         }
 
         /* compute density gradient products */
@@ -194,28 +213,26 @@ inline void Potential::xc_mt_magnetic(Radial_grid<double> const& rgrid,
         auto lapl_rho_dn_lm = laplacian(rho_dn_lm);
 
         /* backward transform Laplacians from Rlm to (theta, phi) */
-        lapl_rho_up_tp = transform(sht_.get(), lapl_rho_up_lm);
-        lapl_rho_dn_tp = transform(sht_.get(), lapl_rho_dn_lm);
+        lapl_rho_up_tp = transform(*sht_, lapl_rho_up_lm);
+        lapl_rho_dn_tp = transform(*sht_, lapl_rho_dn_lm);
     }
 
-    Spheric_function<spatial, double> vsigma_uu_tp;
-    Spheric_function<spatial, double> vsigma_ud_tp;
-    Spheric_function<spatial, double> vsigma_dd_tp;
-    if (is_gga)
-    {
-        vsigma_uu_tp = Spheric_function<spatial, double>(sht_->num_points(), rgrid);
+    Spheric_function<function_domain_t::spatial, double> vsigma_uu_tp;
+    Spheric_function<function_domain_t::spatial, double> vsigma_ud_tp;
+    Spheric_function<function_domain_t::spatial, double> vsigma_dd_tp;
+    if (is_gga) {
+        vsigma_uu_tp = Spheric_function<function_domain_t::spatial, double>(sht_->num_points(), rgrid);
         vsigma_uu_tp.zero();
 
-        vsigma_ud_tp = Spheric_function<spatial, double>(sht_->num_points(), rgrid);
+        vsigma_ud_tp = Spheric_function<function_domain_t::spatial, double>(sht_->num_points(), rgrid);
         vsigma_ud_tp.zero();
 
-        vsigma_dd_tp = Spheric_function<spatial, double>(sht_->num_points(), rgrid);
+        vsigma_dd_tp = Spheric_function<function_domain_t::spatial, double>(sht_->num_points(), rgrid);
         vsigma_dd_tp.zero();
     }
 
     /* loop over XC functionals */
-    for (auto& ixc: xc_func)
-    {
+    for (auto& ixc: xc_func) {
         /* if this is an LDA functional */
         if (ixc.is_lda()) {
             #pragma omp parallel
@@ -281,9 +298,9 @@ inline void Potential::xc_mt_magnetic(Radial_grid<double> const& rgrid,
 
     if (is_gga) {
         /* forward transform vsigma to Rlm */
-        auto vsigma_uu_lm = transform(sht_.get(), vsigma_uu_tp);
-        auto vsigma_ud_lm = transform(sht_.get(), vsigma_ud_tp);
-        auto vsigma_dd_lm = transform(sht_.get(), vsigma_dd_tp);
+        auto vsigma_uu_lm = transform(*sht_, vsigma_uu_tp);
+        auto vsigma_ud_lm = transform(*sht_, vsigma_ud_tp);
+        auto vsigma_dd_lm = transform(*sht_, vsigma_dd_tp);
 
         /* compute gradient of vsgima in spherical harmonics */
         auto grad_vsigma_uu_lm = gradient(vsigma_uu_lm);
@@ -291,14 +308,13 @@ inline void Potential::xc_mt_magnetic(Radial_grid<double> const& rgrid,
         auto grad_vsigma_dd_lm = gradient(vsigma_dd_lm);
 
         /* backward transform gradient from Rlm to (theta, phi) */
-        Spheric_function_gradient<spatial, double> grad_vsigma_uu_tp(sht_->num_points(), rgrid);
-        Spheric_function_gradient<spatial, double> grad_vsigma_ud_tp(sht_->num_points(), rgrid);
-        Spheric_function_gradient<spatial, double> grad_vsigma_dd_tp(sht_->num_points(), rgrid);
-        for (int x = 0; x < 3; x++)
-        {
-            grad_vsigma_uu_tp[x] = transform(sht_.get(), grad_vsigma_uu_lm[x]);
-            grad_vsigma_ud_tp[x] = transform(sht_.get(), grad_vsigma_ud_lm[x]);
-            grad_vsigma_dd_tp[x] = transform(sht_.get(), grad_vsigma_dd_lm[x]);
+        Spheric_vector_function<function_domain_t::spatial, double> grad_vsigma_uu_tp(sht_->num_points(), rgrid);
+        Spheric_vector_function<function_domain_t::spatial, double> grad_vsigma_ud_tp(sht_->num_points(), rgrid);
+        Spheric_vector_function<function_domain_t::spatial, double> grad_vsigma_dd_tp(sht_->num_points(), rgrid);
+        for (int x = 0; x < 3; x++) {
+            grad_vsigma_uu_tp[x] = transform(*sht_, grad_vsigma_uu_lm[x]);
+            grad_vsigma_ud_tp[x] = transform(*sht_, grad_vsigma_ud_lm[x]);
+            grad_vsigma_dd_tp[x] = transform(*sht_, grad_vsigma_dd_lm[x]);
         }
 
         /* compute scalar product of two gradients */
@@ -308,10 +324,8 @@ inline void Potential::xc_mt_magnetic(Radial_grid<double> const& rgrid,
         auto grad_vsigma_ud_grad_rho_dn_tp = grad_vsigma_ud_tp * grad_rho_dn_tp;
 
         /* add remaining terms to Vxc */
-        for (int ir = 0; ir < rgrid.num_points(); ir++)
-        {
-            for (int itp = 0; itp < sht_->num_points(); itp++)
-            {
+        for (int ir = 0; ir < rgrid.num_points(); ir++) {
+            for (int itp = 0; itp < sht_->num_points(); itp++) {
                 vxc_up_tp(itp, ir) -= (2 * grad_vsigma_uu_grad_rho_up_tp(itp, ir) + grad_vsigma_ud_grad_rho_dn_tp(itp, ir));
                 vxc_dn_tp(itp, ir) -= (2 * grad_vsigma_dd_grad_rho_dn_tp(itp, ir) + grad_vsigma_ud_grad_rho_up_tp(itp, ir));
             }
@@ -329,21 +343,21 @@ inline void Potential::xc_mt(Density const& density__)
         int nmtp = unit_cell_.atom(ia).num_mt_points();
 
         /* backward transform density from Rlm to (theta, phi) */
-        auto rho_tp = transform(sht_.get(), density__.rho().f_mt(ialoc));
+        auto rho_tp = transform(*sht_, density__.rho().f_mt(ialoc));
 
         /* backward transform magnetization from Rlm to (theta, phi) */
-        std::vector< Spheric_function<spatial, double> > vecmagtp(ctx_.num_mag_dims());
+        std::vector<Spheric_function<function_domain_t::spatial, double> > vecmagtp(ctx_.num_mag_dims());
         for (int j = 0; j < ctx_.num_mag_dims(); j++) {
-            vecmagtp[j] = transform(sht_.get(), density__.magnetization(j).f_mt(ialoc));
+            vecmagtp[j] = transform(*sht_, density__.magnetization(j).f_mt(ialoc));
         }
 
         /* "up" component of the density */
-        Spheric_function<spectral, double> rho_up_lm;
-        Spheric_function<spatial, double> rho_up_tp(sht_->num_points(), rgrid);
+        Spheric_function<function_domain_t::spectral, double> rho_up_lm;
+        Spheric_function<function_domain_t::spatial, double> rho_up_tp(sht_->num_points(), rgrid);
 
         /* "dn" component of the density */
-        Spheric_function<spectral, double> rho_dn_lm;
-        Spheric_function<spatial, double> rho_dn_tp(sht_->num_points(), rgrid);
+        Spheric_function<function_domain_t::spectral, double> rho_dn_lm;
+        Spheric_function<function_domain_t::spatial, double> rho_dn_tp(sht_->num_points(), rgrid);
 
         /* check if density has negative values */
         double rhomin = 0.0;
@@ -397,18 +411,18 @@ inline void Potential::xc_mt(Density const& density__)
             }
 
             /* transform from (theta, phi) to Rlm */
-            rho_up_lm = transform(sht_.get(), rho_up_tp);
-            rho_dn_lm = transform(sht_.get(), rho_dn_tp);
+            rho_up_lm = transform(*sht_, rho_up_tp);
+            rho_dn_lm = transform(*sht_, rho_dn_tp);
         }
 
-        Spheric_function<spatial, double> exc_tp(sht_->num_points(), rgrid);
-        Spheric_function<spatial, double> vxc_tp(sht_->num_points(), rgrid);
+        Spheric_function<function_domain_t::spatial, double> exc_tp(sht_->num_points(), rgrid);
+        Spheric_function<function_domain_t::spatial, double> vxc_tp(sht_->num_points(), rgrid);
 
         if (ctx_.num_spins() == 1) {
             xc_mt_nonmagnetic(rgrid, xc_func_, density__.rho().f_mt(ialoc), rho_tp, vxc_tp, exc_tp);
         } else {
-            Spheric_function<spatial, double> vxc_up_tp(sht_->num_points(), rgrid);
-            Spheric_function<spatial, double> vxc_dn_tp(sht_->num_points(), rgrid);
+            Spheric_function<function_domain_t::spatial, double> vxc_up_tp(sht_->num_points(), rgrid);
+            Spheric_function<function_domain_t::spatial, double> vxc_dn_tp(sht_->num_points(), rgrid);
 
             xc_mt_magnetic(rgrid, xc_func_, rho_up_lm, rho_up_tp, rho_dn_lm, rho_dn_tp, vxc_up_tp, vxc_dn_tp, exc_tp);
 
@@ -436,7 +450,7 @@ inline void Potential::xc_mt(Density const& density__)
             std::array<int, 3> comp_map = {2, 0, 1};
             /* convert magnetic field back to Rlm */
             for (int j = 0; j < ctx_.num_mag_dims(); j++) {
-                auto bxcrlm = transform(sht_.get(), vecmagtp[j]);
+                auto bxcrlm = transform(*sht_, vecmagtp[j]);
                 for (int ir = 0; ir < nmtp; ir++) {
                     /* add auxiliary magnetic field antiparallel to starting magnetization */
                     bxcrlm(0, ir) -= aux_bf_(j, ia) * ctx_.unit_cell().atom(ia).vector_field()[comp_map[j]];
@@ -448,8 +462,8 @@ inline void Potential::xc_mt(Density const& density__)
         }
 
         /* forward transform from (theta, phi) to Rlm */
-        auto vxcrlm = transform(sht_.get(), vxc_tp);
-        auto excrlm = transform(sht_.get(), exc_tp);
+        auto vxcrlm = transform(*sht_, vxc_tp);
+        auto excrlm = transform(*sht_, exc_tp);
         for (int ir = 0; ir < nmtp; ir++) {
             for (int lm = 0; lm < ctx_.lmmax_pot(); lm++) {
                 xc_potential_->f_mt<index_domain_t::local>(lm, ir, ialoc) = vxcrlm(lm, ir);
@@ -529,8 +543,6 @@ inline void Potential::xc_rg_nonmagnetic(Density const& density__)
     Smooth_periodic_function<double> lapl_rho;
     Smooth_periodic_function<double> grad_rho_grad_rho;
 
-    Smooth_periodic_vector_function<double> vsigma_grad_rho;
-
     Smooth_periodic_function<double> div_vsigma_grad_rho;
 
     if (is_gga) {
@@ -562,9 +574,6 @@ inline void Potential::xc_rg_nonmagnetic(Density const& density__)
                 utils::print_hash("grad_rho_grad_rho", h2);
             }
         }
-        if (!use_2nd_deriv) {
-            vsigma_grad_rho = Smooth_periodic_vector_function<double>(ctx_.fft(), gvp);
-        }
     }
 
     mdarray<double, 1> exc_tmp(num_points);
@@ -618,10 +627,10 @@ inline void Potential::xc_rg_nonmagnetic(Density const& density__)
             TERMINATE("You should not be there since SIRIUS is not compiled with libVDWXC support\n");
 #endif
         } else {
-#pragma omp parallel
+            #pragma omp parallel
             {
                 /* split local size between threads */
-                splindex<block> spl_np_t(num_points, omp_get_num_threads(), omp_get_thread_num());
+                splindex<splindex_t::block> spl_np_t(num_points, omp_get_num_threads(), omp_get_thread_num());
 
                 std::vector<double> exc_t(spl_np_t.local_size());
 
@@ -706,6 +715,8 @@ inline void Potential::xc_rg_nonmagnetic(Density const& density__)
                 vxc_tmp(ir) -= 2 * (vsigma.f_rg(ir) * lapl_rho.f_rg(ir) + grad_vsigma_grad_rho.f_rg(ir));
             }
         } else {
+            Smooth_periodic_vector_function<double> vsigma_grad_rho(ctx_.fft(), gvp);
+
             for (int x: {0, 1, 2}) {
                 for (int ir = 0; ir < num_points; ir++) {
                     vsigma_grad_rho[x].f_rg(ir) = grad_rho[x].f_rg(ir) * vsigma.f_rg(ir);
@@ -714,7 +725,8 @@ inline void Potential::xc_rg_nonmagnetic(Density const& density__)
                 vsigma_grad_rho[x].fft_transform(-1);
             }
             div_vsigma_grad_rho = divergence(vsigma_grad_rho);
-            div_vsigma_grad_rho.fft_transform(-1);
+            /* transform to real space domain */
+            div_vsigma_grad_rho.fft_transform(1);
             for (int ir = 0; ir < num_points; ir++) {
                 vxc_tmp(ir) -= 2 * div_vsigma_grad_rho.f_rg(ir);
             }
@@ -798,8 +810,6 @@ inline void Potential::xc_rg_magnetic(Density const& density__)
 
     Smooth_periodic_vector_function<double> grad_rho_up;
     Smooth_periodic_vector_function<double> grad_rho_dn;
-    Smooth_periodic_function<double> lapl_rho_up;
-    Smooth_periodic_function<double> lapl_rho_dn;
     Smooth_periodic_function<double> grad_rho_up_grad_rho_up;
     Smooth_periodic_function<double> grad_rho_up_grad_rho_dn;
     Smooth_periodic_function<double> grad_rho_dn_grad_rho_dn;
@@ -813,8 +823,6 @@ inline void Potential::xc_rg_magnetic(Density const& density__)
         /* generate pw coeffs of the gradient and laplacian */
         grad_rho_up = gradient(rho_up);
         grad_rho_dn = gradient(rho_dn);
-        lapl_rho_up = laplacian(rho_up);
-        lapl_rho_dn = laplacian(rho_dn);
 
         /* gradient in real space */
         for (int x: {0, 1, 2}) {
@@ -827,21 +835,12 @@ inline void Potential::xc_rg_magnetic(Density const& density__)
         grad_rho_up_grad_rho_dn = dot(grad_rho_up, grad_rho_dn);
         grad_rho_dn_grad_rho_dn = dot(grad_rho_dn, grad_rho_dn);
 
-        /* Laplacian in real space */
-        lapl_rho_up.fft_transform(1);
-        lapl_rho_dn.fft_transform(1);
-
         if (ctx_.control().print_hash_) {
-            auto h1 = lapl_rho_up.hash_f_rg();
-            auto h2 = lapl_rho_dn.hash_f_rg();
-
             auto h3 = grad_rho_up_grad_rho_up.hash_f_rg();
             auto h4 = grad_rho_up_grad_rho_dn.hash_f_rg();
             auto h5 = grad_rho_dn_grad_rho_dn.hash_f_rg();
 
             if (ctx_.comm().rank() == 0) {
-                utils::print_hash("lapl_rho_up", h1);
-                utils::print_hash("lapl_rho_dn", h2);
                 utils::print_hash("grad_rho_up_grad_rho_up", h3);
                 utils::print_hash("grad_rho_up_grad_rho_dn", h4);
                 utils::print_hash("grad_rho_dn_grad_rho_dn", h5);
@@ -904,8 +903,8 @@ inline void Potential::xc_rg_magnetic(Density const& density__)
                 exc_tmp(i) += exc_t[i];
 
                 /* directly add to Vxc available contributions */
-                vxc_up_tmp(i) += vrho_up_t[i] - 2 * vsigma_uu_t[i] * lapl_rho_up.f_rg(i);
-                vxc_dn_tmp(i) += vrho_dn_t[i] - 2 * vsigma_dd_t[i] * lapl_rho_dn.f_rg(i);
+                vxc_up_tmp(i) += vrho_up_t[i];
+                vxc_dn_tmp(i) += vrho_dn_t[i];
 
                 /* save the sigma derivative */
                 vsigma_uu_tmp(i) += vsigma_uu_t[i];
@@ -919,7 +918,7 @@ inline void Potential::xc_rg_magnetic(Density const& density__)
 #pragma omp parallel
             {
                 /* split local size between threads */
-                splindex<block> spl_t(num_points, omp_get_num_threads(), omp_get_thread_num());
+                splindex<splindex_t::block> spl_t(num_points, omp_get_num_threads(), omp_get_thread_num());
 
                 std::vector<double> exc_t(spl_t.local_size());
 
@@ -970,10 +969,9 @@ inline void Potential::xc_rg_magnetic(Density const& density__)
                     for (int i = 0; i < spl_t.local_size(); i++) {
                         /* add Exc contribution */
                         exc_tmp(spl_t[i]) += exc_t[i];
-
                         /* directly add to Vxc available contributions */
-                        vxc_up_tmp(spl_t[i]) += (vrho_up_t[i] - 2 * vsigma_uu_t[i] * lapl_rho_up.f_rg(spl_t[i]) - vsigma_ud_t[i] * lapl_rho_dn.f_rg(spl_t[i]));
-                        vxc_dn_tmp(spl_t[i]) += (vrho_dn_t[i] - 2 * vsigma_dd_t[i] * lapl_rho_dn.f_rg(spl_t[i]) - vsigma_ud_t[i] * lapl_rho_up.f_rg(spl_t[i]));
+                        vxc_up_tmp(spl_t[i]) += vrho_up_t[i];
+                        vxc_dn_tmp(spl_t[i]) += vrho_dn_t[i];
 
                         /* save the sigma derivative */
                         vsigma_uu_tmp(spl_t[i]) += vsigma_uu_t[i];
@@ -989,8 +987,11 @@ inline void Potential::xc_rg_magnetic(Density const& density__)
     if (is_gga) {
         utils::timer t4("sirius::Potential::xc_rg_magnetic|grad2");
         /* gather vsigma */
+        // vsigma_uu: dϵ/dσ↑↑
         Smooth_periodic_function<double> vsigma_uu(ctx_.fft(), ctx_.gvec_partition());
+        // vsigma_ud: dϵ/dσ↑↓
         Smooth_periodic_function<double> vsigma_ud(ctx_.fft(), ctx_.gvec_partition());
+        // vsigma_dd: dϵ/dσ↓↓
         Smooth_periodic_function<double> vsigma_dd(ctx_.fft(), ctx_.gvec_partition());
         for (int ir = 0; ir < num_points; ir++) {
             vsigma_uu.f_rg(ir) = vsigma_uu_tmp[ir];
@@ -998,34 +999,28 @@ inline void Potential::xc_rg_magnetic(Density const& density__)
             vsigma_dd.f_rg(ir) = vsigma_dd_tmp[ir];
         }
 
-        /* forward transform vsigma to plane-wave domain */
-        vsigma_uu.fft_transform(-1);
-        vsigma_ud.fft_transform(-1);
-        vsigma_dd.fft_transform(-1);
-
-        /* gradient of vsigma in plane-wave domain */
-        auto grad_vsigma_uu = gradient(vsigma_uu);
-        auto grad_vsigma_ud = gradient(vsigma_ud);
-        auto grad_vsigma_dd = gradient(vsigma_dd);
-
-        /* backward transform gradient from pw to real space */
+        Smooth_periodic_vector_function<double> up_gradrho_vsigma(ctx_.fft(), ctx_.gvec_partition());
+        Smooth_periodic_vector_function<double> dn_gradrho_vsigma(ctx_.fft(), ctx_.gvec_partition());
         for (int x: {0, 1, 2}) {
-            grad_vsigma_uu[x].fft_transform(1);
-            grad_vsigma_ud[x].fft_transform(1);
-            grad_vsigma_dd[x].fft_transform(1);
+            for(int ir = 0; ir < num_points; ++ir) {
+              up_gradrho_vsigma[x].f_rg(ir) = 2 * grad_rho_up[x].f_rg(ir) * vsigma_uu.f_rg(ir) + grad_rho_dn[x].f_rg(ir) * vsigma_ud.f_rg(ir);
+              dn_gradrho_vsigma[x].f_rg(ir) = 2 * grad_rho_dn[x].f_rg(ir) * vsigma_dd.f_rg(ir) + grad_rho_up[x].f_rg(ir) * vsigma_ud.f_rg(ir);
+            }
+            /* transform to plane wave domain */
+            up_gradrho_vsigma[x].fft_transform(-1);
+            dn_gradrho_vsigma[x].fft_transform(-1);
         }
 
-        /* compute scalar product of two gradients */
-        auto grad_vsigma_uu_grad_rho_up = dot(grad_vsigma_uu, grad_rho_up);
-        auto grad_vsigma_dd_grad_rho_dn = dot(grad_vsigma_dd, grad_rho_dn);
-        auto grad_vsigma_ud_grad_rho_up = dot(grad_vsigma_ud, grad_rho_up);
-        auto grad_vsigma_ud_grad_rho_dn = dot(grad_vsigma_ud, grad_rho_dn);
+        auto div_up_gradrho_vsigma = divergence(up_gradrho_vsigma);
+        div_up_gradrho_vsigma.fft_transform(1);
+        auto div_dn_gradrho_vsigma = divergence(dn_gradrho_vsigma);
+        div_dn_gradrho_vsigma.fft_transform(1);
 
         /* add remaining term to Vxc */
         #pragma omp parallel for
         for (int ir = 0; ir < num_points; ir++) {
-            vxc_up_tmp(ir) -= (2 * grad_vsigma_uu_grad_rho_up.f_rg(ir) + grad_vsigma_ud_grad_rho_dn.f_rg(ir));
-            vxc_dn_tmp(ir) -= (2 * grad_vsigma_dd_grad_rho_dn.f_rg(ir) + grad_vsigma_ud_grad_rho_up.f_rg(ir));
+            vxc_up_tmp(ir) -= div_up_gradrho_vsigma.f_rg(ir);
+            vxc_dn_tmp(ir) -= div_dn_gradrho_vsigma.f_rg(ir);
         }
     }
 
