@@ -49,7 +49,7 @@ void inner_local<double>(memory_t mem__, linalg_t la__, int ispn__, Wave_functio
                            beta__, buf__, ld__, sid__);
         /* subtract one extra G=0 contribution */
         if (comm.rank() == 0) {
-            linalg_t la = is_host_memory(mem__) ? linalg_t::blas : linalg_t::cublas;
+            linalg_t la = is_host_memory(mem__) ? linalg_t::blas : linalg_t::gpublas;
             linalg2(la).ger(m__, n__, &linalg_const<double>::m_one(),
                             reinterpret_cast<double*>(bra__.pw_coeffs(s).prime().at(bra__.preferred_memory_t(), 0, i0__)),
                             2 * bra__.pw_coeffs(s).prime().ld(),
