@@ -89,12 +89,11 @@ class Atom
      *  \f[
      *      D_{\xi \xi'}^{\alpha} = \int V({\bf r}) Q_{\xi \xi'}^{\alpha}({\bf r}) d{\bf r}
      *  \f]
+     *
+     *  The ionic part of the D-operator matrix is added in the D_operator class, when it is initialized.
      */
     mdarray<double, 3> d_mtrx_;
 
-    /// D-operator matrix for spin orbit coupling. see Ref. PRB 71 115106 eq.19
-    mdarray<double_complex, 3> d_mtrx_so_;
-
   public:
     /// Constructor.
     Atom(Atom_type const& type__, vector3d<double> position__, vector3d<double> vector_field__)
@@ -144,10 +143,6 @@ class Atom
             int nbf = type().mt_lo_basis_size();
             d_mtrx_ = mdarray<double, 3>(nbf, nbf, type().parameters().num_mag_dims() + 1, memory_t::host, "Atom::d_mtrx_");
             d_mtrx_.zero();
-            if (type().spin_orbit_coupling()) {
-                d_mtrx_so_ = mdarray<double_complex, 3>(nbf, nbf, type().parameters().num_mag_dims() + 1);
-                d_mtrx_so_.zero();
-            }
         }
     }
 
@@ -178,7 +173,7 @@ class Atom
             TERMINATE("not yet mpi parallel");
         }
 
-        splindex<block> spl_lm(lmmax, comm__.size(), comm__.rank());
+        splindex<splindex_t::block> spl_lm(lmmax, comm__.size(), comm__.rank());
 
         auto l_by_lm = utils::l_by_lm(lmax_pot_);
 
@@ -221,7 +216,7 @@ class Atom
 
         mdarray<double, 1> result(idx_ri.size(1));
 
-        if (pu__ == GPU) {
+        if (pu__ == device_t::GPU) {
 #ifdef __GPU
             auto& rgrid    = type().radial_grid();
             auto& rf_coef  = type().rf_coef();
@@ -277,7 +272,7 @@ class Atom
             TERMINATE_NO_GPU
 #endif
         }
-        if (pu__ == CPU) {
+        if (pu__ == device_t::CPU) {
             utils::timer t1("sirius::Atom::generate_radial_integrals|interp");
             #pragma omp parallel
             {
@@ -594,16 +589,6 @@ class Atom
     {
         return d_mtrx_;
     }
-
-    inline double_complex& d_mtrx_so(int xi1, int xi2, int iv)
-    {
-        return d_mtrx_so_(xi1, xi2, iv);
-    }
-
-    inline double_complex const& d_mtrx_so(int xi1, int xi2, int iv) const
-    {
-        return d_mtrx_so_(xi1, xi2, iv);
-    }
 };
 
 } // namespace
