@@ -33,6 +33,7 @@
 #include "radial_functions_index.hpp"
 #include "basis_functions_index.hpp"
 #include "hubbard_orbitals_descriptor.hpp"
+#include "SHT/sht.hpp"
 
 namespace sirius {
 
@@ -81,6 +82,7 @@ class Atom_type
     std::vector<radial_solution_descriptor_set> aw_descriptors_;
 
     /// List of radial descriptor sets used to construct local orbitals.
+    /** This list defines all local orbitals for a given atom type */
     std::vector<local_orbital_descriptor> lo_descriptors_;
 
     /// Maximum number of AW radial functions across angular momentums.
@@ -89,9 +91,11 @@ class Atom_type
     int offset_lo_{-1}; // TODO: better name
 
     /// Index of radial basis functions.
+    /** This index is used in LAPW to combine APW and local-orbital radial functions */
     radial_functions_index indexr_;
 
     /// Index of atomic basis functions (radial function * spherical harmonic).
+    /** This index is used in LAPW to combine APW and local-orbital muffin-tin functions */
     basis_functions_index indexb_;
 
     /// Radial functions of beta-projectors.
@@ -102,10 +106,19 @@ class Atom_type
         Beta-projectors must be loaded before loading the Q radial functions. */
     mdarray<Spline<double>, 2> q_radial_functions_l_;
 
+    /// Index for the radial hubbard basis functions.
+    radial_functions_index indexr_wfc_;
+
+    /// Index of atomic wavefunctions (radial function * spherical harmonic).
+    basis_functions_index indexb_wfc_;
+
     /// Atomic wave-functions used to setup the initial subspace and to apply U-correction.
     /** This are the chi wave-function in the USPP file. Pairs of [l, chi_l(r)] are stored. */
     std::vector<std::pair<int, Spline<double>>> ps_atomic_wfs_;
 
+    /// List of radial descriptor sets used to construct localized orbitals.
+    std::vector<local_orbital_descriptor> wfc_lo_descriptors_;
+
     /// Total occupancy of the (hubbard) wave functions.
     std::vector<double> ps_atomic_wf_occ_;
 
@@ -160,6 +173,15 @@ class Atom_type
     /// Vector containing all orbitals informations that are relevant for the Hubbard correction.
     std::vector<hubbard_orbital_descriptor> hubbard_orbitals_;
 
+    /// index for the radial hubbard basis functions
+    radial_functions_index hubbard_indexr_;
+
+    /// Index of atomic basis functions (radial function * spherical harmonic).
+    basis_functions_index hubbard_indexb_;
+
+    /// List of radial descriptor sets used to construct hubbard orbitals.
+    std::vector<local_orbital_descriptor> hubbard_lo_descriptors_;
+
     /// Hubbard correction.
     bool hubbard_correction_{false};
 
@@ -167,7 +189,7 @@ class Atom_type
     /** Used in Chebyshev iterative solver as a block-diagonal preconditioner */
     matrix<double_complex> p_mtrx_;
 
-    /// f_coefficients defined in Ref. PRB 71 115106 Eq.9 only
+    /// f_coefficients defined in doi:10.1103/PhysRevB.71.115106 Eq.9 only
     /// valid when SO interactions are on
     mdarray<double_complex, 4> f_coefficients_;
 
@@ -229,7 +251,7 @@ class Atom_type
     /* forbid assignment operator */
     Atom_type& operator=(Atom_type const& src) = delete;
 
-  protected:
+protected:
     /// Radial grid of the muffin-tin sphere.
     Radial_grid<double> radial_grid_;
 
@@ -243,7 +265,7 @@ class Atom_type
     /// Radial grid of a free atom.
     Radial_grid<double> free_atom_radial_grid_;
 
-  public:
+public:
     Atom_type(Simulation_parameters const&                parameters__,
               std::string                                 symbol__,
               std::string                                 name__,
@@ -256,16 +278,16 @@ class Atom_type
         , zn_(zn__)
         , mass_(mass__)
         , atomic_levels_(levels__)
-    {
-    }
+        {
+        }
 
     Atom_type(Simulation_parameters const& parameters__, int id__, std::string label__, std::string file_name__)
         : parameters_(parameters__)
         , id_(id__)
         , label_(label__)
         , file_name_(file_name__)
-    {
-    }
+        {
+        }
 
     Atom_type(Atom_type&& src) = default;
 
@@ -282,7 +304,7 @@ class Atom_type
     inline void set_radial_grid(int num_points__, double const* points__)
     {
         radial_grid_ = Radial_grid_ext<double>(num_points__, points__);
-        if (parameters_.processing_unit() == GPU) {
+        if (parameters_.processing_unit() == device_t::GPU) {
             radial_grid_.copy_to_device();
         }
     }
@@ -360,19 +382,28 @@ class Atom_type
         lo_descriptors_.push_back(lod__);
     }
 
-    inline void add_ps_atomic_wf(int n__, int l__, std::vector<double> f__)
-    {
-        Spline<double> s(radial_grid_, f__);
-        ps_atomic_wfs_.push_back(std::move(std::make_pair(l__, std::move(s))));
-        ps_atomic_wf_level_.push_back(n__);
-    }
-
     inline void add_ps_atomic_wf(int n__, int l__, std::vector<double> f__, double occ_)
     {
+        local_orbital_descriptor lod;
+
         Spline<double> s(radial_grid_, f__);
         ps_atomic_wfs_.push_back(std::move(std::make_pair(l__, std::move(s))));
         ps_atomic_wf_occ_.push_back(occ_);
         ps_atomic_wf_level_.push_back(n__);
+        lod.l = std::abs(l__);
+
+        if (l__ < 0) {
+            lod.total_angular_momentum = lod.l - 0.5;
+        } else {
+            lod.total_angular_momentum = lod.l + 0.5;
+        }
+
+        wfc_lo_descriptors_.push_back(lod);
+    }
+
+    inline void add_ps_atomic_wf(int n__, int l__, std::vector<double> f__)
+    {
+        add_ps_atomic_wf(n__, l__, f__, 0.0);
     }
 
     std::pair<int, Spline<double>> const& ps_atomic_wf(int idx__) const
@@ -589,20 +620,27 @@ class Atom_type
     {
         for (int s = 0; s < static_cast<int>(ps_atomic_wf_level_.size()); s++) {
             if (ps_atomic_wf_level_[s] >= 0) {
-                if ((ps_atomic_wf_level_[s] ==  n) && (std::abs(ps_atomic_wf(s).first) == l)) {
+                auto &e = ps_atomic_wf(s);
+                if ((ps_atomic_wf_level_[s] ==  n) && (std::abs(e.first) == l)) {
                     hubbard_orbital_descriptor hub(n, l, s, occ, J, U, hub_coef__, alpha__, beta__, J0__);
                     hubbard_orbitals_.push_back(std::move(hub));
+                    local_orbital_descriptor lod;
+                    lod.l = std::abs(e.first);
 
+                    if (e.first < 0) {
+                        lod.total_angular_momentum = lod.l - 0.5;
+                    } else {
+                        lod.total_angular_momentum = lod.l + 0.5;
+                    }
                     // we nedd to consider the case where spin orbit
                     // coupling is included. if so is included then we need
                     // to search for its partner with same n, l, but
                     // different j. if not we can stop the for loop
-
+                    hubbard_lo_descriptors_.push_back(lod);
                     if (!this->spin_orbit_coupling_)
                         break;
                 }
             } else {
-
                 auto &wfc = ps_atomic_wfs_[s];
                 // we do a search per angular momentum
                 // we pick the first atomic wave function we
@@ -612,7 +650,9 @@ class Atom_type
                 if (std::abs(wfc.first) == l) {
                     hubbard_orbital_descriptor hub(n, l, s, occ, J, U, hub_coef__, alpha__, beta__, J0__);
                     hubbard_orbitals_.push_back(std::move(hub));
-
+                    local_orbital_descriptor lod;
+                    lod.l = std::abs(ps_atomic_wf(s).first);
+                    hubbard_lo_descriptors_.push_back(lod);
                     if (!this->spin_orbit_coupling_)
                         break;
                 }
@@ -620,7 +660,7 @@ class Atom_type
         }
     }
 
-    inline int number_of_hubbard_channels() const
+    inline int number_of_hubbard_orbitals() const
     {
         return static_cast<int>(hubbard_orbitals_.size());
     }
@@ -838,6 +878,21 @@ class Atom_type
         return indexr_.size();
     }
 
+    inline basis_functions_index const& indexb_wfc() const
+    {
+        return indexb_wfc_;
+    }
+
+    inline basis_functions_index const& hubbard_indexb_wfc() const
+    {
+        return hubbard_indexb_;
+    }
+
+    inline radial_functions_index const& hubbard_indexr() const
+    {
+        return hubbard_indexr_;
+    }
+
     inline void set_symbol(const std::string symbol__)
     {
         symbol_ = symbol__;
@@ -956,6 +1011,8 @@ class Atom_type
         return parameters_;
     }
 
+    /// Set radial grid of the free atom.
+    /** The grid is extended to effective infinity (usually > 100 a.u.) */
     inline void set_free_atom_radial_grid(int num_points__, double const* points__)
     {
         if (num_points__ <= 0) {
@@ -1015,13 +1072,13 @@ class Atom_type
   private:
     void read_hubbard_input();
     void generate_f_coefficients(void);
-    inline double ClebschGordan(const int l, const double j, const double m, const int spin);
-    inline double_complex calculate_U_sigma_m(const int l, const double j, const int mj, const int m, const int sigma);
+    // inline double ClebschGordan(const int l, const double j, const double m, const int spin);
+    // inline double_complex calculate_U_sigma_m(const int l, const double j, const int mj, const int m, const int sigma);
 };
 
-inline void Atom_type::init(int offset_lo__)
-{
-    PROFILE("sirius::Atom_type::init");
+    inline void Atom_type::init(int offset_lo__)
+    {
+        PROFILE("sirius::Atom_type::init");
 
     /* check if the class instance was already initialized */
     if (initialized_) {
@@ -1089,7 +1146,7 @@ inline void Atom_type::init(int offset_lo__)
             lod.l = std::abs(e.first);
 
             /* for spin orbit coupling; we can always do that there is
-               no insidence on the reset when calculations exclude SO */
+               no insidence on the rest when calculations exclude SO */
             if (e.first < 0) {
                 lod.total_angular_momentum = lod.l - 0.5;
             } else {
@@ -1105,6 +1162,17 @@ inline void Atom_type::init(int offset_lo__)
     /* initialize index of muffin-tin basis functions */
     indexb_.init(indexr_);
 
+    /* initialize index for wave functions */
+
+    indexr_wfc_.init(wfc_lo_descriptors_);
+    indexb_wfc_.init(indexr_wfc_);
+
+    if (hubbard_correction_) {
+        /* circus for the hubbard orbitals */
+        hubbard_indexr_.init(hubbard_lo_descriptors_);
+        hubbard_indexb_.init(hubbard_indexr_);
+    }
+
     if (!parameters_.full_potential()) {
         assert(mt_radial_basis_size() == num_beta_radial_functions());
         assert(lmax_beta() == indexr().lmax());
@@ -1192,38 +1260,6 @@ inline void Atom_type::init_free_atom_density(bool smooth)
         /* interpolate at this point near MT radius */
         double R = free_atom_radial_grid_[irmt];
 
-        //std::vector<double> g;
-        //double nel = fourpi * free_atom_density_spline_.integrate(g, 2);
-
-        //std::cout << "number of electrons: " << nel << ", inside MT: " << g[irmt] * fourpi << "\n";
-
-
-        //mdarray<double, 1> b(2);
-        //mdarray<double, 2> A(2, 2);
-        //A(0, 0) = std::pow(R, 2);
-        //A(0, 1) = std::pow(R, 3);
-        //A(1, 0) = 2 * R;
-        //A(1, 1) = 3 * std::pow(R, 2);
-        ////A(0, 0) = 4; //std::pow(R, 1);
-        ////A(0, 1) = std::pow(R, 1);
-        ////A(1, 0) = 0;
-        ////A(1, 1) = 1; //2 * std::pow(R, 1);
-
-        //b(0) = free_atom_density_spline_(irmt);
-        //b(1) = free_atom_density_spline_.deriv(1, irmt);
-
-        //linalg<CPU>::gesv<double>(2, 1, A.at<CPU>(), 2, b.at<CPU>(), 2);
-
-        //////== /* write initial density */
-        //std::stringstream sstr;
-        //sstr << "free_density_" << id_ << ".dat";
-        //FILE* fout = fopen(sstr.str().c_str(), "w");
-
-        //for (int ir = 0; ir < free_atom_radial_grid().num_points(); ir++) {
-        //    fprintf(fout, "%18.12f %18.12f \n", free_atom_radial_grid(ir), free_atom_density_[ir]);
-        //}
-        //fclose(fout);
-
         /* make smooth free atom density inside muffin-tin */
         for (int i = 0; i <= irmt; i++) {
             double x = free_atom_radial_grid(i);
@@ -1318,7 +1354,7 @@ inline void Atom_type::print_info() const
     printf("total number of basis functions  : %i\n", indexb().size());
     printf("number of aw basis functions     : %i\n", indexb().size_aw());
     printf("number of lo basis functions     : %i\n", indexb().size_lo());
-    if (!parameters_.full_potential()) 
+    if (!parameters_.full_potential())
     {
         printf("number of ps wavefunctions       : %i\n", this->num_ps_atomic_wf());
     }
@@ -1573,8 +1609,13 @@ inline void Atom_type::read_pseudo_paw(json const& parser)
 {
     is_paw_ = true;
 
+    auto& header = parser["pseudo_potential"]["header"];
     /* read core energy */
-    paw_core_energy(parser["pseudo_potential"]["header"]["paw_core_energy"]);
+    if (header.count("paw_core_energy")) {
+        paw_core_energy(header["paw_core_energy"]);
+    } else {
+        paw_core_energy(0);
+    }
 
     /* cutoff index */
     int cutoff_radius_index = parser["pseudo_potential"]["header"]["cutoff_radius_index"];
@@ -1663,98 +1704,6 @@ inline void Atom_type::read_input(std::string const& str__)
     read_hubbard_input();
 }
 
-// TODO: this is not an atom type property, move to SHT or utils class
-inline double Atom_type::ClebschGordan(const int l, const double j, const double mj, const int spin)
-{
-    // l : orbital angular momentum
-    // m:  projection of the total angular momentum $m \pm /frac12$
-    // spin: Component of the spinor, 0 up, 1 down
-
-    double CG = 0.0; // Clebsch Gordan coeeficient cf PRB 71, 115106 page 3 first column
-
-    if ((spin != 0) && (spin != 1)) {
-        printf("Error : unkown spin direction\n");
-    }
-
-    const double denom = sqrt(1.0 / (2.0 * l + 1.0));
-
-    if (std::abs(j - l - 0.5) < 1e-8) { // check for j = l + 1/2
-        int m = static_cast<int>(mj - 0.5); // if mj is integer (2 * m), then int m = (mj-1) >> 1;
-        if (spin == 0) {
-            CG = sqrt(l + m + 1.0);
-        }
-        if (spin == 1) {
-            CG = sqrt((l - m));
-        }
-    } else {
-        if (std::abs(j - l + 0.5) < 1e-8) { // check for j = l - 1/2
-            int m = static_cast<int>(mj + 0.5);
-            if (m < (1 - l)) {
-                CG = 0.0;
-            } else {
-                if (spin == 0) {
-                    CG = sqrt(l - m + 1);
-                }
-                if (spin == 1) {
-                    CG = -sqrt(l + m);
-                }
-            }
-        } else {
-            printf("Clebsch gordan coefficients do not exist for this combination of j=%.5lf and l=%d\n", j, l);
-            exit(0);
-        }
-    }
-    return (CG * denom);
-}
-
-// this function computes the U^sigma_{ljm mj} coefficient that
-// rotates the complex spherical harmonics to the real one for the
-// spin orbit case
-
-// mj is normally half integer from -j to j but to avoid computation
-// error it is considered as integer so mj = 2 mj
-
-inline double_complex
-Atom_type::calculate_U_sigma_m(const int l, const double j, const int mj, const int mp, const int sigma)
-{
-
-    if ((sigma != 0) && (sigma != 1)) {
-        printf("SphericalIndex function : unkown spin direction\n");
-        return 0;
-    }
-
-    if (std::abs(j - l - 0.5) < 1e-8) {
-        // j = l + 1/2
-        // m = mj - 1/2
-
-        int m1 = (mj - 1) >> 1;
-        if (sigma == 0) { // up spin
-            if (m1 < -l) { // convention U^s_{mj,m'} = 0
-                return 0.0;
-            } else {// U^s_{mj,mp} =
-                return SHT::rlm_dot_ylm(l, m1, mp);
-            }
-        } else { // down spin
-            if ((m1 + 1) > l) {
-                return 0.0;
-            } else {
-                return SHT::rlm_dot_ylm(l, m1 + 1, mp);
-            }
-        }
-    } else {
-        if (std::abs(j - l + 0.5) < 1e-8) {
-            int m1 = (mj + 1) >> 1;
-            if (sigma == 0) {
-                return SHT::rlm_dot_ylm(l, m1 - 1, mp);
-            } else {
-                return SHT::rlm_dot_ylm(l, m1, mp);
-            }
-        } else {
-            printf("Spherical Index function : l and j are not compatible\n");
-            exit(0);
-        }
-    }
-}
 
 inline void Atom_type::generate_f_coefficients(void)
 {
@@ -1762,7 +1711,7 @@ inline void Atom_type::generate_f_coefficients(void)
 
     // First thing, we need to compute the
     // \f[f^{\sigma\sigma^\prime}_{l,j,m;l\prime,j\prime,m\prime}\f]
-    // They are defined by Eq.9 of Ref PRB 71, 115106
+    // They are defined by Eq.9 of doi:10.1103/PhysRevB.71.115106
     // and correspond to transformations of the
     // spherical harmonics
     if (!this->spin_orbit_coupling()) {
@@ -1796,10 +1745,10 @@ inline void Atom_type::generate_f_coefficients(void)
 
                         int jj1 = static_cast<int>(2.0 * j1 + 1e-8);
                         for (int mj = -jj1; mj <= jj1; mj += 2) {
-                            coef += calculate_U_sigma_m(l1, j1, mj, m1, sigma1) *
-                                    this->ClebschGordan(l1, j1, mj / 2.0, sigma1) *
-                                    std::conj(calculate_U_sigma_m(l2, j2, mj, m2, sigma2)) *
-                                    this->ClebschGordan(l2, j2, mj / 2.0, sigma2);
+                            coef += SHT::calculate_U_sigma_m(l1, j1, mj, m1, sigma1) *
+                                SHT::ClebschGordan(l1, j1, mj / 2.0, sigma1) *
+                                std::conj(SHT::calculate_U_sigma_m(l2, j2, mj, m2, sigma2)) *
+                                SHT::ClebschGordan(l2, j2, mj / 2.0, sigma2);
                         }
                         f_coefficients_(xi1, xi2, sigma1, sigma2) = coef;
                     }
