@@ -31,6 +31,7 @@
 #include "sddk.hpp"
 
 using namespace geometry3d;
+using namespace nlohmann;
 
 namespace sirius {
 
@@ -245,7 +246,7 @@ struct Iterative_solver_input
 
     /// Additional tolerance for empty states.
     /** Setting this variable to 0 will treat empty states with the same tolerance as occupied states. */
-    double empty_states_tolerance_{1e-5};
+    double empty_states_tolerance_{0};
 
     /// Defines the flavour of the iterative solver.
     /** If converge_by_energy is set to 0, then the residuals are estimated by their norm. If converge_by_energy
@@ -302,7 +303,6 @@ struct Iterative_solver_input
  *      "reduce_gvec" : (bool) use reduced G-vector set (reduce_gvec = true) or full set (reduce_gvec = false)
  *      "std_evp_solver_type" : (string) type of eigen-solver for the standard eigen-problem
  *      "gen_evp_solver_type" : (string) type of eigen-solver for the generalized eigen-problem
- *      "electronic_structure_method" : (string) electronic structure method
  *      "processing_unit" : (string) primary processing unit
  *      "fft_mode" : (string) serial or parallel FFT
  *    }
@@ -458,13 +458,13 @@ struct Parameters_input
     double smearing_width_{0.01}; // in Ha
 
     /// Cutoff for plane-waves (for density and potential expansion).
-    double pw_cutoff_{20.0}; // in a.u.^-1
+    double pw_cutoff_{0.0}; // in a.u.^-1
 
     /// Cutoff for augmented-wave functions.
-    double aw_cutoff_{7.0}; // this is R_{MT} * |G+k|_{max}
+    double aw_cutoff_{0.0}; // this is R_{MT} * |G+k|_{max}
 
     /// Cutoff for |G+k| plane-waves.
-    double gk_cutoff_{6.0}; // in a.u.^-1
+    double gk_cutoff_{0.0}; // in a.u.^-1
 
     /// Maximum l for APW functions.
     int lmax_apw_{8};
@@ -523,6 +523,9 @@ struct Parameters_input
     /// Reduction of the auxiliary magnetic field at each SCF step.
     double reduce_aux_bf_{0.0};
 
+    /// Introduce extra charge to the system. Positive charge means extra holes, negative charge - extra electrons.
+    double extra_charge_{0.0};
+
     void read(json const& parser)
     {
         if (parser.count("parameters")) {
@@ -570,6 +573,7 @@ struct Parameters_input
             molecule_       = parser["parameters"].value("molecule", molecule_);
             nn_radius_      = parser["parameters"].value("nn_radius", nn_radius_);
             reduce_aux_bf_  = parser["parameters"].value("reduce_aux_bf", reduce_aux_bf_);
+            extra_charge_   = parser["parameters"].value("extra_charge", extra_charge_);
 
             if (parser["parameters"].count("spin_orbit")) {
                 so_correction_ = parser["parameters"].value("spin_orbit", so_correction_);
@@ -597,7 +601,8 @@ struct Settings_input
     int nprii_aug_{20};
     int nprii_rho_core_{20};
     bool always_update_wf_{true};
-    double mixer_rss_min_{1e-12};
+    double mixer_rss_min_{1e-16};
+    double itsol_tol_min_{1e-13};
     double auto_enu_tol_{0};
     std::string radial_grid_{"exponential, 1.0"};
 
@@ -610,6 +615,7 @@ struct Settings_input
             nprii_rho_core_   = parser["settings"].value("nprii_rho_core", nprii_rho_core_);
             always_update_wf_ = parser["settings"].value("always_update_wf", always_update_wf_);
             mixer_rss_min_    = parser["settings"].value("mixer_rss_min", mixer_rss_min_);
+            itsol_tol_min_    = parser["settings"].value("itsol_tol_min", itsol_tol_min_);
             auto_enu_tol_     = parser["settings"].value("auto_enu_tol", auto_enu_tol_);
             radial_grid_      = parser["settings"].value("radial_grid", radial_grid_);
         }
