@@ -125,7 +125,7 @@ inline std::vector<int> l_by_lm(int lmax__)
             v[lm(l, m)] = l;
         }
     }
-    return std::move(v);
+    return v;
 }
 
 /// Check if file exists.
@@ -310,7 +310,7 @@ inline nlohmann::json read_json_from_file_or_string(std::string const& str__)
 {
     nlohmann::json dict = {};
     if (str__.size() == 0) {
-        return std::move(dict);
+        return dict;
     }
 
     if (str__.find("{") == std::string::npos) { /* this is a file */
@@ -340,7 +340,7 @@ inline nlohmann::json read_json_from_file_or_string(std::string const& str__)
         }
     }
 
-    return std::move(dict);
+    return dict;
 }
 
 /// Get high water mark and resident space size values of a given process.
@@ -429,19 +429,16 @@ inline std::complex<double> conj(std::complex<double> x__)
 }
 
 template <typename T>
-inline T zero_if_not_complex(std::complex<double> x__);
-
-template<>
-inline double zero_if_not_complex<double>(std::complex<double> x__)
-{
-    return 0;
-}
-
-template<>
-inline std::complex<double> zero_if_not_complex<std::complex<double>>(std::complex<double> x__)
+inline T zero_if_not_complex(T x__)
 {
     return x__;
-}
+};
+
+template <typename T>
+inline T zero_if_not_complex(std::complex<T> x__)
+{
+    return 0;
+};
 
 /// Simple random number generator.
 inline uint32_t rnd()
